import os
from datetime import datetime
from typing import Iterator

import pytest

from ... import MercutoClient
from ...ingester.mercuto import MercutoIngester
from ...mocks import mock_mercuto

CAMPBELL_SAMPLE_FILE = os.path.join(os.path.dirname(__file__), 'resources', 'campbell-sample-file.dat')


@pytest.fixture
def mock_client() -> Iterator[MercutoClient]:
    with mock_mercuto():
        yield MercutoClient()


def test_samples_upload(mock_client: MercutoClient) -> None:
    tenant = mock_client.identity().create_tenant('Test Tenant', 'T123456789')
    project = mock_client.core().create_project(
        'test_project',
        'R123456789',
        'Test Project',
        tenant.code,
        timezone='UTC'
    )

    channel1 = mock_client.data().create_channel(
        project=project.code,
        label="VWu_1"
    )
    channel2 = mock_client.data().create_channel(
        project=project.code,
        label="VWu_2"
    )

    ingester = MercutoIngester(
        project_code=project.code,
        api_key='test_api_key',
        timezone='UTC',
    )

    assert ingester.process_file(CAMPBELL_SAMPLE_FILE)

    data = mock_client.data().load_secondary_samples(
        channels=[channel1.code, channel2.code],
        start_time=datetime.fromisoformat('2023-12-07T00:01:00+00:00'),
        end_time=datetime.fromisoformat('2023-12-07T00:04:00+00:00'),
    )
    assert len(data) == 8



import argparse
import logging
import logging.handlers
import os
import signal
import sys
import time
from pathlib import Path
from typing import Any, Callable, Optional, TypeVar
from urllib.parse import ParseResult, urlparse

import schedule

from ..util import get_free_space_excluding_files
from .backup import get_backup_handler
from .ftp import simple_ftp_server
from .mercuto import MercutoIngester
from .pid_file import PidFile
from .processor import FileProcessor

logger = logging.getLogger(__name__)


T = TypeVar('T')


def call_and_log_error(func: Callable[[], T]) -> T | None:
    """
    Call a function and log any exceptions that occur.
    """
    try:
        return func()
    except Exception:
        logging.exception(f"Error in {func.__name__}")
        return None


class Status:
    """
    Status class to handle running state of the ingester.
    """

    def __init__(self):
        self.running = True

    def stop(self, code: Any, frame: Any):
        self.running = False
        print("Stopping")

    def is_running(self):
        return self.running


def launch_mercuto_ingester(
    project: str,
    api_key: str,
    hostname: str = 'https://api.rockfieldcloud.com.au',
    verify_ssl: bool = True,
    pid_file: Optional[Path] = None,
    workdir: Optional[str] = '~/.mercuto-ingester',
    verbose: bool = False,
    logfile: Optional[str] = None,
    directory: Optional[str] = None,
    target_free_space_mb: Optional[float] = None,
    max_files: Optional[int] = None,
    mapping: Optional[str] = None,
    clean: bool = False,
    ftp_server_username: str = 'logger',
    ftp_server_password: str = 'password',
    ftp_server_port: int = 2121,
    ftp_server_rename: bool = True,
    max_attempts: int = 1000,
    backup_location: Optional[list[ParseResult]] = None,
    timezone: Optional[str] = None,
    camera: Optional[str] = None
):

    if backup_location is None:
        backup_location = []

    with PidFile(pid_file):
        if workdir is None:
            workdir = os.path.join(os.path.expanduser('~'), ".mercuto-ingester")
        elif workdir.startswith("~"):
            workdir = os.path.expanduser(workdir)
        else:
            workdir = workdir
            if not os.path.exists(workdir):
                raise ValueError(f"Work directory {workdir} does not exist")

        os.makedirs(workdir, exist_ok=True)

        if verbose:
            level = logging.DEBUG
        else:
            level = logging.INFO

        handlers: list[logging.Handler] = []
        handlers.append(logging.StreamHandler(sys.stderr))

        if logfile is not None:
            logfile = logfile
        else:
            logfile = os.path.join(workdir, 'log.txt')
        handlers.append(logging.handlers.RotatingFileHandler(
            logfile, maxBytes=1000000, backupCount=3))

        logging.basicConfig(format='[PID %(process)d] %(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
                            datefmt='%d/%m/%Y %H:%M:%S',
                            level=level,
                            handlers=handlers)

        if directory is None:
            buffer_directory = os.path.join(workdir, "buffered-files")
        else:
            buffer_directory = directory
        os.makedirs(buffer_directory, exist_ok=True)

        ftp_dir = os.path.join(workdir, 'temp-ftp-data')
        os.makedirs(ftp_dir, exist_ok=True)

        if target_free_space_mb is None and max_files is None:
            target_free_space_mb = get_free_space_excluding_files(buffer_directory) * 0.25 // (1024 * 1024)  # Convert to MB
            logging.info(f"Target remaining free space set to {target_free_space_mb} MB based on available disk space.")

            if target_free_space_mb <= 1:
                raise ValueError("Not enough free space on the buffer partition to set a reasonable target free space. "
                                 "Please specify either a larger buffer directory or set the target free space or max files manually.")

        logger.info(f"Using work directory: {workdir}")

        database_path = os.path.join(workdir, "buffer.db")
        if clean and os.path.exists(database_path):
            logging.info(f"Dropping existing database at {database_path}")
            os.remove(database_path)

        ingester = MercutoIngester(
            project_code=project,
            api_key=api_key,
            hostname=hostname,
            verify_ssl=verify_ssl,
            timezone=timezone,
            camera_code=camera
        )

        if mapping is not None:
            import json
            with open(mapping, 'r') as f:
                mapping = json.load(f)
            if not isinstance(mapping, dict):
                raise ValueError(f"Mapping file {mapping} must contain a JSON object")
            ingester.update_mapping(mapping)

        pre_processing_handlers: list[Callable[[str], bool]] = [get_backup_handler(loc) for loc in backup_location]
        processor_callbacks: list[Callable[[str], bool]] = [ingester.process_file]
        post_processing_handlers: list[Callable[[str], bool]] = []

        all_handlers = pre_processing_handlers + processor_callbacks + post_processing_handlers

        processor = FileProcessor(
            buffer_dir=buffer_directory,
            db_path=database_path,
            process_callback=lambda filename: all(handler(filename) for handler in all_handlers),
            max_attempts=max_attempts,
            target_free_space_mb=target_free_space_mb,
            max_files=max_files)

        processor.scan_existing_files()

        with simple_ftp_server(directory=buffer_directory,
                               username=ftp_server_username, password=ftp_server_password, port=ftp_server_port,
                               callback=processor.add_file_to_db, rename=ftp_server_rename,
                               workdir=ftp_dir):
            call_and_log_error(ingester.ping)
            schedule.every(60).seconds.do(call_and_log_error, ingester.ping)  # type: ignore[attr-defined]
            schedule.every(5).seconds.do(call_and_log_error, processor.process_next_file)  # type: ignore[attr-defined]
            schedule.every(2).minutes.do(call_and_log_error, processor.cleanup_old_files)  # type: ignore[attr-defined]

            status = Status()
            signal.signal(signal.SIGTERM, status.stop)

            while status.is_running():
                schedule.run_pending()
                sleep_period = schedule.idle_seconds()
                if sleep_period is None or sleep_period < 0:
                    sleep_period = 0
                # We need to wake up to handle ctrl-c etc
                if sleep_period > 1:
                    sleep_period = 1
                time.sleep(sleep_period)

            logger.warning("Shutting Down...")


def main():
    parser = argparse.ArgumentParser(description='Mercuto Ingester CLI')
    parser.add_argument('-p', '--project', type=str,
                        required=True, help='Mercuto project code')
    parser.add_argument('-k', '--api-key', type=str,
                        required=True, help='API key for Mercuto')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Enable verbose output')
    parser.add_argument('-d', '--directory', type=str,
                        help='Directory to store ingested files. Default is a directory called `buffered-files` in the workdir.')
    parser.add_argument('-s', '--target-free-space-mb', type=int,
                        help='Size in MB for total amount of remaining free space to keep available. \
                            Default is 25% of the available disk space on the buffer partition excluding the directory itself', default=None)
    parser.add_argument('--max-files', type=int,
                        help='Maximum number of files to keep in the buffer. Default is to use the size param.', default=None)
    parser.add_argument('--max-attempts', type=int,
                        help='Maximum number of attempts to process a file before giving up. Default is 1000.',
                        default=1000)
    parser.add_argument('--workdir', type=str,
                        help='Working directory for the ingester. Default is ~/.mercuto-ingester',)
    parser.add_argument('--logfile', type=str,
                        help='Log file path. No logs written if not provided. Maximum of 4 log files of 1MB each will be kept.\
                            Default is log.txt in the workdir.')
    parser.add_argument('--mapping', type=str,
                        help='Path to a JSON file with channel label to channel code mapping.\
                            If not provided, the ingester will try to detect the channels from the project.',
                        default=None)
    parser.add_argument('--hostname', type=str,
                        help='Hostname to use for the Mercuto server. Default is "https://api.rockfieldcloud.com.au".',
                        default='https://api.rockfieldcloud.com.au')
    parser.add_argument('--clean',
                        help='Drop the database before starting. This will not remove any buffer files and will rescan them on startup.',
                        action='store_true')
    parser.add_argument('--username', type=str,
                        help='Username for the FTP server. Default is "logger".',
                        default='logger')
    parser.add_argument('--password', type=str,
                        help='Password for the FTP server. Default is "password".',
                        default='password')
    parser.add_argument('--port', type=int,
                        help='Port for the FTP server. Default is 2121.',
                        default=2121)
    parser.add_argument('--no-rename', action='store_true',
                        help='Add the current timestamp to the end of the files received via FTP. \
                        This is useful to avoid overwriting files with the same name.')
    parser.add_argument('-i', '--insecure', action="store_true",
                        help='Disable SSL verification',
                        default=False)
    parser.add_argument('-b', '--backup-location', action="append",
                        help='Backup location to store ingested files. Must be a valid URL, e.g. scp://user@host/path. '
                        'Can be specified multiple times.',
                        type=urlparse)
    parser.add_argument('-e', '--pid-file', help='Ths location to create the PID file', type=Path, default=None)
    parser.add_argument('--timezone', type=str,
                        help='Timezone to use for data uploads (e.g. "Australia/Melbourne"). \
                        If not provided, no timezone will be sent on uploads. \
                        Only needed if data files do not contain timezone information (E.g. Campbell Scientific loggers).',
                        default=None)
    parser.add_argument('--camera', type=str,
                        help='Camera code to associate with image uploads. If not provided, image files will error on upload.',
                        default=None)

    args = parser.parse_args()

    launch_mercuto_ingester(
        project=args.project,
        api_key=args.api_key,
        verify_ssl=not args.insecure,
        pid_file=args.pid_file,
        workdir=args.workdir,
        verbose=args.verbose,
        logfile=args.logfile,
        directory=args.directory,
        target_free_space_mb=args.target_free_space_mb,
        max_files=args.max_files,
        mapping=args.mapping,
        clean=args.clean,
        ftp_server_username=args.username,
        ftp_server_password=args.password,
        ftp_server_port=args.port,
        ftp_server_rename=not args.no_rename,
        max_attempts=args.max_attempts,
        backup_location=args.backup_location,
        hostname=args.hostname,
        timezone=args.timezone,
        camera=args.camera
    )


if __name__ == '__main__':
    main()



import json
from typing import TypedDict, TypeVar


class ResourceTypes:
    WILDCARD = '*'

    class Mercuto:
        """
        Resource types available for ServiceTypes.MERCUTO
        """
        PROJECT = 'project'
        SYSTEM = 'system'
        WILDCARD = '*'

    class Identity:
        """
        Resource types available for ServiceTypes.IDENTITY
        """
        TENANT = 'tenant'
        WILDCARD = '*'

    class Notifications:
        """
        Resource types available for ServiceTypes.NOTIFICATIONS
        """
        CONTACT_GROUP = 'contact-group'
        WILDCARD = '*'


class AllowedActions:
    WILDCARD = '*'

    class Mercuto:
        """
        Actions available for ServiceTypes.MERCUTO
        """
        WILDCARD = '*'
        VIEW_PROJECT = 'MERCUTO:VIEW_PROJECT'
        MANAGE_PROJECT = 'MERCUTO:MANAGE_PROJECT'
        EDIT_PROJECT = 'MERCUTO:EDIT_PROJECT'
        UPLOAD_DATA = 'MERCUTO:UPLOAD_DATA'
        EDIT_SYSTEM = 'MERCUTO:EDIT_SYSTEM'

    class Identity:
        """
        Actions available for ServiceTypes.IDENTITY
        """
        WILDCARD = '*'
        VIEW_TENANT = 'IDENTITY:VIEW_TENANT'
        MANAGE_TENANT = 'IDENTITY:MANAGE_TENANT'
        EDIT_TENANT = 'IDENTITY:EDIT_TENANT'
        VIEW_USER_DETAILED_INFO = 'IDENTITY:VIEW_USER_DETAILED_INFO'
        CREATE_NEW_TENANTS = 'IDENTITY:CREATE_NEW_TENANTS'

    class Notifications:
        """
        Actions available for ServiceTypes.NOTIFICATIONS
        """
        WILDCARD = '*'
        ISSUE_NOTIFICATION = 'NOTIFICATIONS:ISSUE_NOTIFICATION'


class ServiceTypes:
    class Mercuto:
        Name = 'mercuto'
        ResourceTypes = ResourceTypes.Mercuto
        AllowedActions = AllowedActions.Mercuto

    class Identity:
        Name = 'identity'
        ResourceTypes = ResourceTypes.Identity
        AllowedActions = AllowedActions.Identity

    class Notifications:
        Name = 'notifications'
        ResourceTypes = ResourceTypes.Notifications
        AllowedActions = AllowedActions.Notifications

    IDENTITY = Identity.Name
    MERCUTO = Mercuto.Name
    NOTIFICATIONS = Notifications.Name
    WILDCARD = '*'


_T = TypeVar('_T', bound='AclPolicyBuilder')


class AclPolicyBuilder:
    class _AclEntryType(TypedDict):
        action: str
        resource: str

    class _AclPolicyType(TypedDict):
        version: int
        permissions: list['AclPolicyBuilder._AclEntryType']

    def __init__(self) -> None:
        self._permissions: list[AclPolicyBuilder._AclEntryType] = []

    def allow(self: _T, action: str, resource: str) -> _T:
        self._permissions.append({
            'action': action,
            'resource': resource
        })
        return self

    def allow_all(self: _T, action: str) -> _T:
        self.allow(action, f"mrn:{ServiceTypes.WILDCARD}:{ResourceTypes.WILDCARD}/{ResourceTypes.WILDCARD}")
        return self

    def allow_project(self: _T, action: str, project_code: str) -> _T:
        self.allow(action, f"mrn:{ServiceTypes.MERCUTO}:{ResourceTypes.Mercuto.PROJECT}/{project_code}")
        return self

    def allow_tenant(self: _T, action: str, tenant_code: str) -> _T:
        self.allow(action, f"mrn:{ServiceTypes.IDENTITY}:{ResourceTypes.Identity.TENANT}/{tenant_code}")
        return self

    def as_string(self) -> str:
        return json.dumps(self.as_dict())

    def as_dict(self) -> _AclPolicyType:
        return {
            'version': 1,
            'permissions': self._permissions
        }



import os
import tempfile
from pathlib import Path
from threading import Thread
from typing import Iterator, TypedDict
from urllib.parse import urlparse

import pyftpdlib.authorizers  # type: ignore[import-untyped]
import pyftpdlib.handlers  # type: ignore[import-untyped]
import pyftpdlib.servers  # type: ignore[import-untyped]
import pytest

from ...ingester.backup import FileBackup, FTPBackup


def test_file_backup():
    with tempfile.TemporaryDirectory() as temp_dir:
        uri = Path(temp_dir).as_uri()
        bak = FileBackup(urlparse(uri))
        assert bak.process_file(__file__)


def test_file_backup_does_not_exist_create_it():
    with tempfile.TemporaryDirectory() as temp_dir:
        test_dir = Path(temp_dir) / "test_dir"
        uri = test_dir.as_uri() + "?create=true"
        assert not test_dir.exists()
        bak = FileBackup(urlparse(uri))
        assert bak.process_file(__file__)
        dest = test_dir / Path(__file__).name
        assert test_dir.exists()
        assert dest.exists()


def test_file_backup_does_not_exist():
    uri = (Path(__file__).parent / "I_DO_NOT_EXIST").as_uri()
    with pytest.raises(ValueError, match="backup path does not exist"):
        FileBackup(urlparse(uri))


class TestFTPServerConfig(TypedDict):
    host: str
    port: int
    user: str
    passwd: str
    ftp_root: str


@pytest.fixture(scope="function")
def mock_ftp_server() -> Iterator[TestFTPServerConfig]:
    # Create a temporary directory to act as the FTP root
    temp_dir = tempfile.TemporaryDirectory()
    ftp_root = temp_dir.name

    # Set up user authentication
    authorizer = pyftpdlib.authorizers.DummyAuthorizer()
    authorizer.add_user("user", "12345", ftp_root, perm="elradfmwMT")

    # Set up FTP handler
    handler = pyftpdlib.handlers.FTPHandler
    handler.authorizer = authorizer

    # Start the FTP server in a separate thread
    server = pyftpdlib.servers.FTPServer(("127.0.0.1", 0), handler)
    ip, port = server.socket.getsockname()

    server_thread = Thread(target=server.serve_forever)
    server_thread.daemon = True
    server_thread.start()

    yield {
        "host": ip,
        "port": port,
        "user": "user",
        "passwd": "12345",
        "ftp_root": ftp_root
    }

    server.close_all()
    temp_dir.cleanup()


def test_ftp_backup(mock_ftp_server: TestFTPServerConfig):
    url = f"ftp://{mock_ftp_server['user']}:{mock_ftp_server['passwd']}@{mock_ftp_server['host']}:{mock_ftp_server['port']}/my_dir?create=true"
    os.makedirs(Path(mock_ftp_server['ftp_root']) / "my_dir", exist_ok=True)
    bak = FTPBackup(urlparse(url))
    bak.process_file(__file__)
    assert bak.process_file(__file__)
    dest = Path(mock_ftp_server['ftp_root']) / "my_dir" / Path(__file__).name
    assert dest.exists()


def test_ftp_backup_fails_if_dir_not_exists(mock_ftp_server: TestFTPServerConfig):
    url = f"ftp://{mock_ftp_server['user']}:{mock_ftp_server['passwd']}@{mock_ftp_server['host']}:{mock_ftp_server['port']}/my_dir"
    bak = FTPBackup(urlparse(url))
    assert not bak.process_file(__file__)
    dest = Path(mock_ftp_server['ftp_root']) / "my_dir" / Path(__file__).name
    assert not dest.exists()



import contextlib
import json as json_stdlib
import logging
import os
import time
from typing import (Any, Iterator, Literal, Mapping, Optional, Protocol, Type,
                    TypeVar)

import requests
import requests.cookies

from ._authentication import (IAuthenticationMethod,
                              create_authentication_method)
from .exceptions import MercutoClientException, MercutoHTTPException
from .modules.alerts import MercutoAlertService
from .modules.core import MercutoCoreService
from .modules.data import MercutoDataService
from .modules.fatigue import MercutoFatigueService
from .modules.identity import MercutoIdentityService
from .modules.media import MercutoMediaService
from .modules.notifications import MercutoNotificationService
from .modules.reports import MercutoReportService

logger = logging.getLogger(__name__)


class _ModuleBase(Protocol):
    def __init__(self, client: 'MercutoClient', *args: Any, **kwargs: Any) -> None:
        pass


_T = TypeVar('_T', bound=_ModuleBase)


class MercutoClient:
    def __init__(self, url: Optional[str] = None, verify_ssl: bool = True, active_session: Optional[requests.Session] = None) -> None:
        if url is None:
            url = os.environ.get('MERCUTO_API_URL', 'https://api.rockfieldcloud.com.au')
        assert isinstance(url, str)

        if url.endswith('/'):
            url = url[:-1]

        if verify_ssl and not url.startswith('https://'):
            raise ValueError(f'Url must be https, is {url}')

        self._url = url
        self.verify_ssl = verify_ssl

        if active_session is None:
            self._current_session = requests.Session()
        else:
            self._current_session = active_session

        self._auth_method: Optional[IAuthenticationMethod] = None
        self._cookies = requests.cookies.RequestsCookieJar()

        self._modules: dict[str, _ModuleBase] = {}

    def url(self) -> str:
        return self._url

    def credentials_key(self) -> str:
        """
        Generate a unique key that identifies the current credentials set.
        """
        if self._auth_method is None:
            raise MercutoClientException("No credentials set")
        return self._auth_method.unique_key()

    def setverify_ssl(self, verify_ssl: bool) -> None:
        self.verify_ssl = verify_ssl

    def copy(self) -> 'MercutoClient':
        return MercutoClient(self._url, self.verify_ssl, self._current_session)

    @contextlib.contextmanager
    def as_credentials(self, api_key: Optional[str] = None,
                       service_token: Optional[str] = None,
                       bearer_token: Optional[str] = None,
                       headers: Optional[Mapping[str, str]] = None) -> Iterator['MercutoClient']:
        """
        Same as .connect(), but as a context manager. Will automatically logout when exiting the context.
        """
        # TODO: We are passing the current session along to re-use connections for speed. Will this cause security issues?
        other = MercutoClient(self._url, self.verify_ssl, self._current_session)
        try:
            yield other.connect(api_key=api_key, service_token=service_token, bearer_token=bearer_token, headers=headers)
        finally:
            other.logout()

    def connect(self, *, api_key: Optional[str] = None,
                service_token: Optional[str] = None,
                bearer_token: Optional[str] = None,
                headers: Optional[Mapping[str, str]] = None) -> 'MercutoClient':
        """
        Attempt to connect using any available method.
        if api_key is provided, use the api_key.
        if service_token is provided, use the service_token.
        if headers is provided, attempt to extract either api_key or service_token from given header set.
            headers should be a dictionary of headers that would be sent in a request. Useful for using existing authenation mechanism for forwarding.

        """
        authentication = create_authentication_method(api_key=api_key, service_token=service_token, bearer_token=bearer_token, headers=headers)
        self.login(authentication)
        return self

    def _update_headers(self, headers: dict[str, str]) -> dict[str, str]:
        base: dict[str, str] = {}

        if self._auth_method is not None:
            self._auth_method.update_header(base)
        base.update(headers)
        return base

    def session(self) -> requests.Session:
        return self._current_session

    def request(self, url: str, method: Literal['GET', 'POST', 'PUT', 'DELETE', 'PATCH', 'HEAD', 'OPTIONS'],
                params: Optional[dict[str, Any]] = None,
                json: Optional[dict[str, Any]] = None,
                raise_for_status: bool = True,
                **kwargs: Any) -> requests.Response:
        """
        Make an HTTP request to the Mercuto API.
        :param url: The URL path (relative to the base API URL) to make the request to.
        :param method: The HTTP method to use (e.g., 'GET', 'POST', etc.).
        :param params: Optional dictionary of query parameters to include in the request.
        :param json: Optional dictionary to send as a JSON payload in the request body.
        :param raise_for_status: Whether to raise an exception for HTTP error responses.
        :param kwargs: Additional keyword arguments to pass to the requests method.
        :return: The HTTP response object.
        """
        return self._http_request(url, method, params=params, json=json, raise_for_status=raise_for_status, **kwargs)

    def _http_request(self, url: str, method: str,
                      params: Optional[dict[str, Any]] = None,
                      json: Optional[dict[str, Any]] = None,
                      raise_for_status: bool = True,
                      **kwargs: Any) -> requests.Response:
        if url.startswith('/'):
            url = url[1:]
        full_url = f"{self._url}/{url}"

        if 'timeout' not in kwargs:
            kwargs['timeout'] = 10
        kwargs['headers'] = self._update_headers(kwargs.get('headers', {}))

        if 'verify' not in kwargs:
            kwargs['verify'] = self.verify_ssl

        if 'cookies' not in kwargs:
            kwargs['cookies'] = self._cookies

        # Custom parsing json to support NAN
        if json is not None and kwargs.get('data') is None:
            kwargs['data'] = json_stdlib.dumps(json, allow_nan=True)
            kwargs['headers']['Content-Type'] = 'application/json'
            json = None

        start = time.time()
        resp = self._current_session.request(method, full_url, params=params, json=json, **kwargs)
        duration = time.time() - start
        logger.debug("Made request to %s %s in %.2f seconds (code=%s)", method, full_url, duration, resp.status_code)
        if raise_for_status and not resp.ok:
            try:
                error_json = resp.json()
            except Exception:
                raise MercutoHTTPException(resp.text, resp.status_code)
            else:
                if 'detail' in error_json and isinstance(error_json['detail'], str):
                    raise MercutoHTTPException(error_json['detail'], resp.status_code)
                else:
                    raise MercutoHTTPException(resp.text, resp.status_code)
        resp.cookies.update(self._cookies)
        return resp

    def _add_and_fetch_module(self, name: str, module: Type[_T]) -> _T:
        if name not in self._modules:
            self._modules[name] = module(self)
        return self._modules[name]  # type: ignore

    def identity(self) -> 'MercutoIdentityService':
        return self._add_and_fetch_module('identity', MercutoIdentityService)

    def fatigue(self) -> 'MercutoFatigueService':
        return self._add_and_fetch_module('fatigue', MercutoFatigueService)

    def data(self) -> 'MercutoDataService':
        return self._add_and_fetch_module('data', MercutoDataService)

    def core(self) -> 'MercutoCoreService':
        return self._add_and_fetch_module('core', MercutoCoreService)

    def media(self) -> 'MercutoMediaService':
        return self._add_and_fetch_module('media', MercutoMediaService)

    def reports(self) -> 'MercutoReportService':
        return self._add_and_fetch_module('reports', MercutoReportService)

    def notifications(self) -> 'MercutoNotificationService':
        return self._add_and_fetch_module('notifications', MercutoNotificationService)

    def alerts(self) -> 'MercutoAlertService':
        return self._add_and_fetch_module('alerts', MercutoAlertService)

    def login(self, authentication: IAuthenticationMethod) -> None:
        self._auth_method = authentication

    def logout(self) -> None:
        self._auth_method = None

    def is_logged_in(self) -> bool:
        return self._auth_method is not None



from typing import TYPE_CHECKING, Optional

from pydantic import TypeAdapter

if TYPE_CHECKING:
    from ..client import MercutoClient

from . import PayloadType
from ._util import BaseModel


class PermissionGroup(BaseModel):
    tenant: str
    code: str
    label: str
    acl_policy: str


class Tenant(BaseModel):
    code: str
    name: str
    description: str
    logo_url: Optional[str] = None


class HiddenUserAPIKey(BaseModel):
    code: str
    description: str
    last_used: Optional[str]
    custom_policy: Optional[str] = None


class UserDetails(BaseModel):
    code: str
    username: Optional[str] = None
    email_address: Optional[str] = None
    mobile_number: Optional[str] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    api_keys: list[HiddenUserAPIKey] = []


class User(BaseModel):
    code: str
    username: Optional[str] = None
    description: str
    tenant: str
    permission_group: str


class CurrentUser(BaseModel):
    code: str
    username: Optional[str] = None
    description: str
    tenant: Tenant
    permission_group: PermissionGroup
    current_permission_policy: str


class VisibleUserAPIKey(BaseModel):
    code: str
    new_api_key: str
    description: str
    custom_policy: Optional[str]


class VerifyMyPermissions(BaseModel):
    user: Optional[str]
    acl_policy: str


class Healthcheck(BaseModel):
    status: str


# --- TypeAdapters for lists ---
_PermissionGrouplistAdapter = TypeAdapter(list[PermissionGroup])
_TenantlistAdapter = TypeAdapter(list[Tenant])
_UserlistAdapter = TypeAdapter(list[User])
_HiddenUserAPIKeylistAdapter = TypeAdapter(list[HiddenUserAPIKey])


class MercutoIdentityService:
    def __init__(self, client: 'MercutoClient', path: str = '/identity') -> None:
        self._client = client
        self._path = path

    def healthcheck(self) -> Healthcheck:
        r = self._client.request(f"{self._path}/healthcheck", "GET")
        return Healthcheck.model_validate_json(r.text)

    # --- Verify routes ---

    def get_my_permissions(self) -> VerifyMyPermissions:
        r = self._client.request(f"{self._path}/verify/me", "GET")
        return VerifyMyPermissions.model_validate_json(r.text)

    # --- User routes ---

    def list_users(self, tenant: Optional[str] = None) -> list[User]:
        params: PayloadType = {}
        if tenant is not None:
            params["tenant"] = tenant
        r = self._client.request(f"{self._path}/users", "GET", params=params)
        return _UserlistAdapter.validate_json(r.text)

    def create_user(
        self,
        username: str,
        tenant: str,
        description: str,
        group: str,
        default_password: Optional[str] = None
    ) -> User:
        payload: PayloadType = {
            "username": username,
            "tenant_code": tenant,
            "description": description,
            "group_code": group,
            "default_password": default_password,
        }
        r = self._client.request(f"{self._path}/users", "PUT", json=payload)
        return User.model_validate_json(r.text)

    def get_current_user(self) -> CurrentUser:
        r = self._client.request(f"{self._path}/users/me", "GET")
        return CurrentUser.model_validate_json(r.text)

    def get_user(self, code: str) -> User:
        r = self._client.request(f"{self._path}/users/{code}", "GET")
        return User.model_validate_json(r.text)

    def delete_user(self, code: str) -> None:
        self._client.request(f"{self._path}/users/{code}", "DELETE")

    def edit_user(
        self,
        code: str,
        description: str,
        group: str
    ) -> User:
        payload: PayloadType = {
            "description": description,
            "group_code": group,
        }
        r = self._client.request(f"{self._path}/users/{code}", "PATCH", json=payload)
        return User.model_validate_json(r.text)

    def get_user_details(self, code: str) -> UserDetails:
        r = self._client.request(f"{self._path}/users/{code}/details", "GET")
        return UserDetails.model_validate_json(r.text)

    def set_user_details(
        self,
        code: str,
        email_address: Optional[str] = None,
        mobile_number: Optional[str] = None,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None
    ) -> UserDetails:
        payload: PayloadType = {
            "email_address": email_address,
            "mobile_number": mobile_number,
            "first_name": first_name,
            "last_name": last_name,
        }
        r = self._client.request(f"{self._path}/users/{code}/details", "PATCH", json=payload)
        return UserDetails.model_validate_json(r.text)

    def get_user_api_keys(self, user: str) -> list[HiddenUserAPIKey]:
        r = self._client.request(f"{self._path}/users/{user}/api_keys", "GET")
        return _HiddenUserAPIKeylistAdapter.validate_json(r.text)

    def generate_api_key_for_user(
        self,
        user: str,
        description: str,
        custom_policy: Optional[str] = None
    ) -> VisibleUserAPIKey:
        payload: PayloadType = {
            "description": description,
            "custom_policy": custom_policy,
        }
        r = self._client.request(f"{self._path}/users/{user}/api_keys", "POST", json=payload)
        return VisibleUserAPIKey.model_validate_json(r.text)

    def delete_api_key(self, user: str, key_code: str) -> None:
        self._client.request(f"{self._path}/users/{user}/api_keys/{key_code}", "DELETE")

    # --- Tenants routes ---

    def list_tenants(self) -> list[Tenant]:
        r = self._client.request(f"{self._path}/tenants", "GET")
        return _TenantlistAdapter.validate_json(r.text)

    def get_tenant(self, code: str) -> Tenant:
        r = self._client.request(f"{self._path}/tenants/{code}", "GET")
        return Tenant.model_validate_json(r.text)

    def create_tenant(
        self,
        name: str,
        description: str,
        logo_url: Optional[str] = None
    ) -> Tenant:
        payload: PayloadType = {
            "name": name,
            "description": description,
            "logo_url": logo_url,
        }
        r = self._client.request(f"{self._path}/tenants", "PUT", json=payload)
        return Tenant.model_validate_json(r.text)

    # --- Permission Groups routes ---

    def get_permission_groups(self, tenant: Optional[str] = None) -> list[PermissionGroup]:
        params: PayloadType = {}
        if tenant is not None:
            params["tenant"] = tenant
        r = self._client.request(f"{self._path}/permissions", "GET", params=params)
        return _PermissionGrouplistAdapter.validate_json(r.text)

    def create_permission_group(
        self,
        tenant: str,
        label: str,
        acl_policy: str
    ) -> PermissionGroup:
        payload: PayloadType = {
            "tenant": tenant,
            "label": label,
            "acl_policy": acl_policy,
        }
        r = self._client.request(f"{self._path}/permissions", "PUT", json=payload)
        return PermissionGroup.model_validate_json(r.text)

    def get_permission_group(self, group: str) -> PermissionGroup:
        r = self._client.request(f"{self._path}/permissions/{group}", "GET")
        return PermissionGroup.model_validate_json(r.text)

    def delete_permission_group(self, group: str) -> None:
        self._client.request(f"{self._path}/permissions/{group}", "DELETE")

    def modify_permission_group(
        self,
        group: str,
        label: str,
        acl_policy: str
    ) -> None:
        payload: PayloadType = {
            "label": label,
            "acl_policy": acl_policy,
        }
        self._client.request(f"{self._path}/permissions/{group}", "PATCH", json=payload)



from datetime import datetime
from typing import TYPE_CHECKING, Literal, Optional

from pydantic import Field, TypeAdapter

if TYPE_CHECKING:
    from ..client import MercutoClient

from . import PayloadType
from ._util import BaseModel


class RainflowConfiguration(BaseModel):
    project: str
    max_bins: int
    bin_size: float
    multiplier: float
    reservoir_adjustment: bool
    sources: list[str]


class FatigueConnection(BaseModel):
    project: str
    code: str
    label: str
    multiplier: float
    c_d: float
    m: float
    s_0: float
    bs7608_failure_probability: Optional[float]
    bs7608_detail_category: Optional[str]
    initial_date: datetime
    initial_damage: float
    sources: list[str]


class ConnectionRemnantCapacity(BaseModel):
    connection: FatigueConnection
    remaining_life_years: float = Field(description="Remaining life of the connection in years")
    total_damage: float = Field(description="Total damage accumulated in the connection up to the 'end_time' specified")


class Healthcheck(BaseModel):
    status: str


_RainflowConfigurationlistAdapter = TypeAdapter(list[RainflowConfiguration])
_FatigueConnectionlistAdapter = TypeAdapter(list[FatigueConnection])
_ConnectionRemnantCapacitylistAdapter = TypeAdapter(list[ConnectionRemnantCapacity])


class MercutoFatigueService:
    def __init__(self, client: 'MercutoClient', path: str = '/fatigue') -> None:
        self._client = client
        self._path = path

    def healthcheck(self) -> Healthcheck:
        r = self._client.request(f"{self._path}/healthcheck", "GET")
        return Healthcheck.model_validate_json(r.text)

    # --- Rainflow routes ---

    def list_rainflow_config(self, project: str) -> list[RainflowConfiguration]:
        params: PayloadType = {"project": project}
        r = self._client.request(f"{self._path}/rainflow/setup", "GET", params=params)
        return _RainflowConfigurationlistAdapter.validate_json(r.text)

    def setup_rainflow(
        self,
        project: str,
        max_bins: int,
        bin_size: float,
        multiplier: float,
        reservoir_adjustment: bool,
        sources: list[str]
    ) -> RainflowConfiguration:
        payload: PayloadType = {
            "project": project,
            "max_bins": max_bins,
            "bin_size": bin_size,
            "multiplier": multiplier,
            "reservoir_adjustment": reservoir_adjustment,
            "sources": sources,
        }
        r = self._client.request(f"{self._path}/rainflow/setup", "PUT", json=payload)
        return RainflowConfiguration.model_validate_json(r.text)

    def get_cycle_counts(
        self, project: str, start_time: datetime, end_time: datetime
    ) -> bytes:
        params: PayloadType = {
            "project": project,
            "start_time": start_time.isoformat(),
            "end_time": end_time.isoformat(),
        }
        r = self._client.request(
            f"{self._path}/rainflow/cycle_counts", "GET", params=params, stream=True
        )
        return r.content

    def delete_cycle_counts(
        self, project: str, start_time: datetime, end_time: datetime, ignore_if_not_configured: bool = False
    ) -> None:
        params: PayloadType = {
            "project": project,
            "start_time": start_time.isoformat(),
            "end_time": end_time.isoformat(),
            "ignore_if_not_configured": ignore_if_not_configured,
        }
        self._client.request(
            f"{self._path}/rainflow/cycle_counts", "DELETE", params=params
        )

    def calculate_cycle_counts(
        self,
        project: str,
        event: str,
        presigned_url: str,
        mime_type: Literal['application/feather'],
        url_expiry: Optional[datetime] = None,
        ignore_if_not_configured: bool = False
    ) -> None:
        payload: PayloadType = {
            "project": project,
            "event": event,
            "presigned_url": presigned_url,
            "mime_type": mime_type,
        }
        if url_expiry is not None:
            payload["url_expiry"] = url_expiry.isoformat()
        params = {"ignore_if_not_configured": ignore_if_not_configured}
        self._client.request(
            f"{self._path}/rainflow/cycle_counts/calculate", "PUT", json=payload, params=params
        )

    # --- Fatigue Connections routes ---

    def get_connections(self, project: str) -> list[FatigueConnection]:
        params: PayloadType = {"project": project}
        r = self._client.request(f"{self._path}/connections", "GET", params=params)
        return _FatigueConnectionlistAdapter.validate_json(r.text)

    def add_connection(
        self,
        project: str,
        label: str,
        multiplier: float,
        c_d: float,
        m: float,
        s_0: float,
        bs7608_failure_probability: float,
        bs7608_detail_category: str,
        initial_date: datetime,
        initial_damage: float,
        sources: list[str]
    ) -> FatigueConnection:
        payload: PayloadType = {
            "project": project,
            "label": label,
            "multiplier": multiplier,
            "c_d": c_d,
            "m": m,
            "s_0": s_0,
            "bs7608_failure_probability": bs7608_failure_probability,
            "bs7608_detail_category": bs7608_detail_category,
            "initial_date": initial_date.isoformat(),
            "initial_damage": initial_damage,
            "sources": sources,
        }
        r = self._client.request(f"{self._path}/connections", "PUT", json=payload)
        return FatigueConnection.model_validate_json(r.text)

    def delete_connection(self, connection_code: str) -> None:
        self._client.request(f"{self._path}/connections/{connection_code}", "DELETE")

    # --- Connection Data routes ---

    def get_connection_remnant_capacity(
        self, project: str, start_time: datetime, end_time: datetime
    ) -> list[ConnectionRemnantCapacity]:
        params: PayloadType = {
            "project": project,
            "start_time": start_time.isoformat(),
            "end_time": end_time.isoformat(),
        }
        r = self._client.request(
            f"{self._path}/connection_data/remnant-capacity", "GET", params=params
        )
        return _ConnectionRemnantCapacitylistAdapter.validate_json(r.text)



from ... import MercutoClient


def test_add_and_list_users(client: MercutoClient) -> None:
    client.identity().create_user("test_user", "test-tenant", "Description", "Test Group")
    users = client.identity().list_users()
    assert len(users) == 1
    assert "test_user" in [user.username for user in users]



import logging
import mimetypes
import os
import uuid
from datetime import datetime, timedelta, timezone
from typing import Optional

from ..client import MercutoClient
from ..exceptions import MercutoHTTPException
from ..modules.media import (Camera, CameraTrigger, CameraType, Image,
                             MercutoMediaService, Video)
from ._utility import EnforceOverridesMeta, create_data_url

logger = logging.getLogger(__name__)


class MockMercutoMediaService(MercutoMediaService, metaclass=EnforceOverridesMeta):
    def __init__(self, client: 'MercutoClient'):
        super().__init__(client=client)
        self._cameras: dict[str, Camera] = {}
        self._events: dict[str, Image] = {}
        self._videos: dict[str, Video] = {}

    def list_cameras(self, project: str) -> list[Camera]:
        return [cam for cam in self._cameras.values() if cam.project == project]

    def get_camera(self, camera_code: str) -> Camera:
        if camera_code not in self._cameras:
            raise MercutoHTTPException(f"Camera not found: {camera_code}", 404)
        return self._cameras[camera_code]

    def create_camera(self, project: str,
                      label: str,
                      triggers: list[CameraTrigger],
                      encode_timestamp: bool = True,
                      encode_blur: bool = False,
                      blur_steps: Optional[int] = None,
                      blur_sigma: Optional[float] = None,
                      tunnel_address: Optional[str] = None,
                      tunnel_port: Optional[int] = None,
                      tunnel_username: Optional[str] = None,
                      tunnel_password: Optional[str] = None,
                      tunnel_key: Optional[str] = None,
                      camera_ip: Optional[str] = None,
                      camera_port: Optional[int] = None,
                      camera_username: Optional[str] = None,
                      camera_password: Optional[str] = None,
                      camera_serial: Optional[str] = None,
                      camera_type: CameraType = 'DIRECT_RTSP',
                      rtsp_url: Optional[str] = None,) -> Camera:
        code = str(uuid.uuid4())
        camera = Camera(code=code, project=project, label=label, triggers=triggers, encode_timestamp=encode_timestamp,
                        encode_blur=encode_blur, blur_steps=blur_steps, blur_sigma=blur_sigma,
                        tunnel_address=tunnel_address, tunnel_port=tunnel_port, tunnel_username=tunnel_username,
                        tunnel_password=tunnel_password, tunnel_key=tunnel_key, camera_ip=camera_ip,
                        camera_port=camera_port, camera_username=camera_username, camera_password=camera_password,
                        camera_serial=camera_serial, camera_type=camera_type, rtsp_url=rtsp_url)
        self._cameras[code] = camera
        return camera

    def upload_image(self, filename: str, project: str,
                     camera: Optional[str] = None,
                     timestamp: Optional[datetime] = None,
                     event: Optional[str] = None,
                     filedata: Optional[bytes] = None) -> Image:
        code = str(uuid.uuid4())
        mime_type, _ = mimetypes.guess_type(filename, strict=False)
        if mime_type is None:
            raise ValueError("Could not determine MIME type for file")
        if mime_type not in ['image/jpeg', 'image/png', 'image/gif', 'image/bmp', 'image/tiff']:
            raise MercutoHTTPException(f"Unsupported image MIME type: {mime_type}", 400)
        if filedata is not None:
            data = filedata
        else:
            with open(filename, 'rb') as f:
                data = f.read()
        data_url = create_data_url(mime_type, data)
        image = Image(code=code, project=project, camera=camera, timestamp=timestamp, event=event, access_url=data_url,
                      mime_type=mime_type, access_expires=datetime.now(timezone.utc) + timedelta(days=365),
                      size_bytes=len(data), name=os.path.basename(filename))
        self._events[code] = image
        return image

    def list_images(self, project: str,
                    camera: Optional[str] = None,
                    event: Optional[str] = None,
                    start_time: Optional[datetime] = None,
                    end_time: Optional[datetime] = None,
                    limit: int = 10,
                    offset: int = 0,
                    ascending: bool = True) -> list[Image]:
        results = [img for img in self._events.values() if img.project == project]
        if camera:
            results = [img for img in results if img.camera == camera]
        if event:
            results = [img for img in results if img.event == event]
        if start_time:
            results = [img for img in results if img.timestamp and img.timestamp >= start_time]
        if end_time:
            results = [img for img in results if img.timestamp and img.timestamp <= end_time]
        results.sort(key=lambda img: img.timestamp or datetime.min, reverse=not ascending)
        return results[offset:offset + limit]

    def get_image(self, image_code: str) -> Image:
        try:
            return self._events[image_code]
        except KeyError:
            raise MercutoHTTPException(f"Image not found: {image_code}", 404)

    def delete_image(self, image_code: str) -> None:
        try:
            del self._events[image_code]
        except KeyError:
            raise MercutoHTTPException(f"Image not found: {image_code}", 404)

    def upload_video(self, filename: str, project: str, start_time: datetime, end_time: datetime,
                     camera: str | None = None, event: str | None = None) -> str:
        code = str(uuid.uuid4())
        mime_type, _ = mimetypes.guess_type(filename, strict=False)
        if mime_type is None:
            raise ValueError("Could not determine MIME type for file")
        with open(filename, 'rb') as f:
            data = f.read()
            data_url = create_data_url(mime_type, data)
        video = Video(code=code, project=project, camera=camera, start_time=start_time, end_time=end_time, event=event,
                      access_url=data_url, mime_type=mime_type,
                      access_expires=datetime.now(timezone.utc) + timedelta(days=365),
                      size_bytes=len(data), name=os.path.basename(filename))
        self._videos[code] = video
        return code

    def list_videos(self, project: str,
                    camera: Optional[str] = None,
                    event: Optional[str] = None,
                    start_time: Optional[datetime] = None,
                    end_time: Optional[datetime] = None,
                    limit: int = 10,
                    offset: int = 0,
                    ascending: bool = True) -> list[Video]:
        results = [vid for vid in self._videos.values() if vid.project == project]
        if camera:
            results = [vid for vid in results if vid.camera == camera]
        if event:
            results = [vid for vid in results if vid.event == event]
        if start_time:
            results = [vid for vid in results if vid.start_time and vid.start_time >= start_time]
        if end_time:
            results = [vid for vid in results if vid.end_time and vid.end_time <= end_time]
        results.sort(key=lambda vid: vid.start_time or datetime.min, reverse=not ascending)
        return results[offset:offset + limit]

    def get_video(self, video_code: str) -> Video:
        try:
            return self._videos[video_code]
        except KeyError:
            raise MercutoHTTPException(f"Video not found: {video_code}", 404)



import math
import os
import tempfile

import pytest

from ...ingester.parsers import (detect_parser, parse_campbell_file,
                                 parse_worldsensing_compact_file,
                                 parse_worldsensing_standard_file)

RESOURCES_DIR = os.path.join(os.path.dirname(__file__), "resources")


def test_worldsensing_compacted_parser():
    file = os.path.join(RESOURCES_DIR, "worldsensing-compacted-sample-file.dat")
    mapper = {
        "channel1": "12345678",
        "channel2": "abcdefgh",
    }
    samples = parse_worldsensing_compact_file(file, mapper)
    assert len(samples) == 4
    assert samples[0].channel == "12345678"
    assert math.isclose(samples[0].value, -10)
    assert samples[0].timestamp.isoformat() == '2025-05-20T15:00:00'

    assert samples[1].channel == "abcdefgh"
    assert math.isclose(samples[1].value, 5)
    assert samples[0].timestamp.isoformat() == '2025-05-20T15:00:00'

    assert samples[2].channel == "12345678"
    assert math.isclose(samples[2].value, -12)
    assert samples[2].timestamp.isoformat() == '2025-05-20T16:00:00'

    assert samples[3].channel == "abcdefgh"
    assert math.isclose(samples[3].value, 10)
    assert samples[3].timestamp.isoformat() == '2025-05-20T16:00:00'


def test_worldsensing_standard_parser():
    file = os.path.join(RESOURCES_DIR, "worldsensing-standard-sample-file.csv")
    mapper = {
        "AtmPressure-85544-in-mbar": "12345678",
        "freqSqInDigit-85544-VW-Ch1": "abcdefgh",
    }
    samples = parse_worldsensing_standard_file(file, mapper)
    assert len(samples) == 10
    assert samples[0].channel == "12345678"
    assert math.isclose(samples[0].value, 930.5)
    assert samples[0].timestamp.isoformat() == '2024-04-15T12:35:00'

    assert samples[1].channel == "abcdefgh"
    assert math.isclose(samples[1].value, 726.810811024)
    assert samples[1].timestamp.isoformat() == '2024-04-15T12:35:00'

    assert samples[8].channel == "12345678"
    assert math.isclose(samples[8].value, 930.4)
    assert samples[8].timestamp.isoformat() == '2024-04-15T12:39:00'

    assert samples[9].channel == "abcdefgh"
    assert math.isclose(samples[9].value, 726.841502500)
    assert samples[9].timestamp.isoformat() == '2024-04-15T12:39:00'


def test_campbells_parser():
    file = os.path.join(RESOURCES_DIR, "campbell-sample-file.dat")
    mapper = {
        "VWu_1": "aaaaaaaa",
        "VWu_2": "bbbbbbbb",
        "Therm(1)": "cccccccc",
        "Therm(2)": "dddddddd",
        "Diag_Max(1)": "eeeeeeee",
        "Diag_Max(2)": "ffffffff",
    }
    samples = parse_campbell_file(file, mapper)
    assert len(samples) == 6*4

    assert samples[0].channel == "aaaaaaaa"
    assert math.isclose(samples[0].value, 1234.5)
    assert samples[0].timestamp.isoformat() == '2023-12-07T00:01:00'

    for i in range(1, 6):
        assert samples[i].channel == list(mapper.values())[i]
        assert math.isnan(samples[i].value)
        assert samples[i].timestamp.isoformat() == '2023-12-07T00:01:00'

    assert samples[6].channel == "aaaaaaaa"
    assert math.isclose(samples[6].value, 1234.5)
    assert samples[6].timestamp.isoformat() == '2023-12-07T00:02:00'
    for i in range(7, 12):
        assert samples[i].channel == list(mapper.values())[i-6]
        assert math.isnan(samples[i].value)
        assert samples[i].timestamp.isoformat() == '2023-12-07T00:02:00'

    assert samples[12].channel == "aaaaaaaa"
    assert math.isclose(samples[12].value, 1234.5)
    assert samples[12].timestamp.isoformat() == '2023-12-07T00:03:00'
    assert samples[13].channel == "bbbbbbbb"
    assert math.isclose(samples[13].value, 1234.5)
    assert samples[13].timestamp.isoformat() == '2023-12-07T00:03:00'

    for i in range(15, 17):
        assert samples[i].channel == list(mapper.values())[i-12]
        assert math.isnan(samples[i].value)
        assert samples[i].timestamp.isoformat() == '2023-12-07T00:03:00'
    assert samples[17].channel == "ffffffff"
    assert math.isclose(samples[17].value, 1537)
    assert samples[17].timestamp.isoformat() == '2023-12-07T00:03:00'

    assert samples[18].channel == "aaaaaaaa"
    assert math.isclose(samples[18].value, 1234.5)
    assert samples[18].timestamp.isoformat() == '2023-12-07T00:04:00'
    assert samples[19].channel == "bbbbbbbb"
    assert math.isclose(samples[19].value, 1234.5)
    assert samples[19].timestamp.isoformat() == '2023-12-07T00:04:00'
    assert samples[20].channel == "cccccccc"
    assert math.isclose(samples[20].value, 27.5)
    assert samples[20].timestamp.isoformat() == '2023-12-07T00:04:00'
    assert samples[21].channel == "dddddddd"
    assert math.isclose(samples[21].value, 25)
    assert samples[21].timestamp.isoformat() == '2023-12-07T00:04:00'
    assert samples[22].channel == "eeeeeeee"
    assert math.isclose(samples[22].value, 255)
    assert samples[22].timestamp.isoformat() == '2023-12-07T00:04:00'
    assert samples[23].channel == "ffffffff"
    assert math.isclose(samples[23].value, 0)
    assert samples[23].timestamp.isoformat() == '2023-12-07T00:04:00'


def test_detect_file_type():
    compacted_file = os.path.join(RESOURCES_DIR, "worldsensing-compacted-sample-file.dat")
    standard_file = os.path.join(RESOURCES_DIR, "worldsensing-standard-sample-file.csv")
    campbell_file = os.path.join(RESOURCES_DIR, "campbell-sample-file.dat")

    assert detect_parser(compacted_file) == parse_worldsensing_compact_file
    assert detect_parser(standard_file) == parse_worldsensing_standard_file
    assert detect_parser(campbell_file) == parse_campbell_file

    # Test with an unknown file type
    with tempfile.TemporaryDirectory() as dir:
        unknown_file = os.path.join(dir, "unknown-file.txt")
        with open(unknown_file, "w") as f:
            f.write("This is an unknown file format.")

        with pytest.raises(ValueError):
            detect_parser(unknown_file)



from typing import TYPE_CHECKING, Literal, Optional

from pydantic import TypeAdapter

if TYPE_CHECKING:
    from ..client import MercutoClient

from ._util import BaseModel

ContactMethod = Literal['EMAIL', 'SMS']


class ContactGroup(BaseModel):
    code: str
    project: str
    label: str
    users: dict[str, list[ContactMethod]]


class NotificationAttachment(BaseModel):
    """
    Attachment to include in the notification.
    """
    filename: str
    presigned_url: str  # Presigned URL to fetch the attachment from.
    mime_type: str


class Healthcheck(BaseModel):
    status: str


# --- TypeAdapters for lists ---
_ContactGroupListAdapter = TypeAdapter(list[ContactGroup])


class MercutoNotificationService:
    def __init__(self, client: 'MercutoClient', path: str = '/notifications') -> None:
        self._client = client
        self._path = path

    def healthcheck(self) -> Healthcheck:
        r = self._client.request(f"{self._path}/healthcheck", "GET")
        return Healthcheck.model_validate_json(r.text)

    def list_contact_groups(self, project: str) -> list[ContactGroup]:
        r = self._client.request(f"{self._path}/contact-groups", "GET", params={"project": project})
        return _ContactGroupListAdapter.validate_json(r.text)

    def get_contact_group(self, code: str) -> ContactGroup:
        r = self._client.request(f"{self._path}/contact-groups/{code}", "GET")
        return ContactGroup.model_validate_json(r.text)

    def create_contact_group(self, project: str, label: str, users: dict[str, list[ContactMethod]]) -> ContactGroup:
        r = self._client.request(f"{self._path}/contact-groups", "PUT", json={
            "project": project,
            "label": label,
            "users": users
        })
        return ContactGroup.model_validate_json(r.text)

    def issue_notification(self, contact_group: str, subject: str, html: str,
                           alternative_plaintext: Optional[str] = None,
                           attachments: Optional[list[NotificationAttachment]] = None,
                           unsubscribe_placeholder_text: Optional[str] = None) -> None:
        """
        Issue a notification to all contacts within a contact group, based on their contact preferences.

        :param contact_group: The code of the contact group to send the notification to.
        :param subject: The subject of the notification (i.e. email subject line).
        :param html: The HTML content of the notification.
        :param alternative_plaintext: Optional plaintext alternative for the notification.
            Alternative plaintext is used for SMS notifications and for email clients that do not support HTML.
        :param attachments: Optional list of attachments to include in the notification.
            Only applicable for email notifications.
        :param unsubscribe_placeholder_text: Optional placeholder text for unsubscribe links.
            Any text matching this placeholder will be replaced with an unsubscribe link for email notifications.
        :return: None
        """

        self._client.request(f"{self._path}/contact-groups/{contact_group}/notify", "POST", json={
            "subject": subject,
            "html": html,
            "alternative_plaintext": alternative_plaintext,
            "attachments": [attachment.model_dump() for attachment in attachments] if attachments else [],
            "unsubscribe_placeholder_text": unsubscribe_placeholder_text
        })
        return



import logging
import uuid
from datetime import datetime, timezone
from typing import Optional

from pydantic import BaseModel

from ..client import MercutoClient
from ..exceptions import MercutoHTTPException
from ..modules.notifications import (ContactGroup, ContactMethod,
                                     MercutoNotificationService,
                                     NotificationAttachment)
from ._utility import EnforceOverridesMeta

logger = logging.getLogger(__name__)


class IssuedNotification(BaseModel):
    contact_group: ContactGroup
    subject: str
    html: str
    alternative_plaintext: Optional[str] = None
    attachments: Optional[list[NotificationAttachment]] = None
    unsubscribe_placeholder_text: Optional[str] = None

    issued_on: datetime


class MockMercutoNotificationService(MercutoNotificationService, metaclass=EnforceOverridesMeta):
    def __init__(self, client: 'MercutoClient'):
        super().__init__(client=client)
        self.contact_groups: dict[str, ContactGroup] = {}
        self.issued_notifications: list[IssuedNotification] = []

    def list_contact_groups(self, project: str) -> list[ContactGroup]:
        return [group for group in self.contact_groups.values() if group.project == project]

    def get_contact_group(self, code: str) -> ContactGroup:
        if code not in self.contact_groups:
            raise MercutoHTTPException(
                f"Contact group with code {code} not found", 404)
        return self.contact_groups[code]

    def create_contact_group(self, project: str, label: str, users: dict[str, list[ContactMethod]]) -> ContactGroup:
        code = str(uuid.uuid4())
        contact_group = ContactGroup(
            code=code, project=project, label=label, users=users)
        self.contact_groups[code] = contact_group
        return contact_group

    def issue_notification(self, contact_group: str, subject: str, html: str,
                           alternative_plaintext: Optional[str] = None,
                           attachments: Optional[list[NotificationAttachment]] = None,
                           unsubscribe_placeholder_text: Optional[str] = None) -> None:
        group = self.get_contact_group(contact_group)
        issued_notification = IssuedNotification(
            contact_group=group,
            subject=subject,
            html=html,
            alternative_plaintext=alternative_plaintext,
            attachments=attachments,
            unsubscribe_placeholder_text=unsubscribe_placeholder_text,
            issued_on=datetime.now(timezone.utc)
        )
        self.issued_notifications.append(issued_notification)



from datetime import datetime
from typing import (TYPE_CHECKING, BinaryIO, Literal, Optional, Protocol,
                    TypedDict)

from pydantic import AwareDatetime, TypeAdapter

from ..exceptions import MercutoHTTPException
from . import PayloadType
from ._util import BaseModel

if TYPE_CHECKING:
    from ..client import MercutoClient


class ReportConfiguration(BaseModel):
    code: str
    project: str
    label: str
    revision: str
    schedule: Optional[str] = None
    contact_group: Optional[str] = None
    last_scheduled: Optional[AwareDatetime] = None
    custom_policy: Optional[str] = None


ReportLogStatus = Literal['PENDING', 'IN_PROGRESS', 'COMPLETED', 'FAILED']


class ReportLog(BaseModel):
    code: str
    report_configuration: str
    scheduled_start: Optional[AwareDatetime]
    actual_start: AwareDatetime
    actual_finish: Optional[AwareDatetime]
    status: ReportLogStatus
    message: Optional[str]
    access_url: Optional[str]
    mime_type: Optional[str]
    filename: Optional[str]


class ReportSourceCodeRevision(BaseModel):
    code: str
    project: Optional[str]
    revision_date: AwareDatetime
    description: str
    source_code_url: str


class Healthcheck(BaseModel):
    status: str


_ReportConfigurationListAdapter = TypeAdapter(list[ReportConfiguration])
_ReportLogListAdapter = TypeAdapter(list[ReportLog])

"""
The below types are used for defining report generation functions.
They are provided for type-checking and helpers for users writing custom reports.
"""


class ReportHandlerResultLike(Protocol):
    filename: str
    mime_type: str
    data: bytes


class ReportHandlerResult(BaseModel):
    filename: str
    mime_type: str
    data: bytes


class HandlerRequest(TypedDict):
    timestamp: datetime


class HandlerContext(TypedDict):
    service_token: str
    project_code: str
    report_code: str
    log_code: str
    client: 'MercutoClient'


class ReportHandler(Protocol):
    def __call__(self,
                 request: 'HandlerRequest',
                 context: 'HandlerContext') -> 'ReportHandlerResultLike':
        ...


class MercutoReportService:
    def __init__(self, client: 'MercutoClient', path: str = '/reports') -> None:
        self._client = client
        self._path = path

    def healthcheck(self) -> Healthcheck:
        r = self._client.request(f"{self._path}/healthcheck", "GET")
        return Healthcheck.model_validate_json(r.text)

    def list_report_configurations(self, project: str) -> list['ReportConfiguration']:
        """
        List scheduled reports for a specific project.
        """
        params: PayloadType = {
            'project': project
        }
        r = self._client.request(
            f'{self._path}/configurations', 'GET', params=params)
        return _ReportConfigurationListAdapter.validate_json(r.text)

    def create_report_configuration(self, project: str, label: str, schedule: str, revision: str,
                                    contact_group: Optional[str] = None, custom_policy: Optional[str] = None) -> ReportConfiguration:
        """
        Create a new scheduled report using the provided source code revision.
        """
        json: PayloadType = {
            'project': project,
            'label': label,
            'schedule': schedule,
            'revision': revision,
        }
        if contact_group is not None:
            json['contact_group'] = contact_group
        if custom_policy is not None:
            json['custom_policy'] = custom_policy
        r = self._client.request(
            f'{self._path}/configurations', 'PUT', json=json)
        return ReportConfiguration.model_validate_json(r.text)

    def generate_report(self, report: str, timestamp: datetime, mark_as_scheduled: bool = False) -> ReportLog:
        """
        Trigger generation of a scheduled report for a specific timestamp.
        """
        r = self._client.request(f'{self._path}/configurations/{report}/generate', 'POST', json={
            'timestamp': timestamp.isoformat(),
            'mark_as_scheduled': mark_as_scheduled
        })
        return ReportLog.model_validate_json(r.text)

    def list_report_logs(self, project: str, report: Optional[str] = None) -> list[ReportLog]:
        """
        List report log entries for a specific project.
        """
        params: PayloadType = {
            'project': project
        }
        if report is not None:
            params['configuration'] = report
        r = self._client.request(
            f'{self._path}/logs', 'GET', params=params)
        return _ReportLogListAdapter.validate_json(r.text)

    def get_report_log(self, log: str) -> ReportLog:
        """
        Get a specific report log entry.
        """
        r = self._client.request(
            f'{self._path}/logs/{log}', 'GET')
        return ReportLog.model_validate_json(r.text)

    def create_report_revision(self, revision_date: datetime,
                               description: str,
                               project: Optional[str],
                               source_code: BinaryIO) -> ReportSourceCodeRevision:
        """
        Create a new report source code revision.

        A report should be a python file that defines a function called `generate_report`
        that takes two arguments: `request` and `context`, and returns an object with
        `filename`, `mime_type`, and `data` attributes. It can also be a package with __init__.py
        defining the `generate_report` function.

        You can use the `mercuto_client.modules.reports.ReportHandler` protocol
        to type hint your report function. Example:
        ```python
        from mercuto_client.modules.reports import ReportHandler, HandlerRequest, HandlerContext, ReportHandlerResult
        def generate_report(request: HandlerRequest, context: HandlerContext) -> ReportHandlerResult:
            # Your report generation logic here
            return ReportHandlerResult(
                filename="report.pdf",
                mime_type="application/pdf",
                data=b"PDF binary data here"
            )
        ```
        The request parameter contains information about the report generation request,
        and the context parameter provides access to the Mercuto client and metadata about
        the report being generated. The MercutoClient provided in the context can be used
        to fetch any additional data required for the report. It will be authenticated
        using a service token with VIEW_PROJECT permission and VIEW_TENANT permission.

        Params:
            project (str): The project code.
            revision_date (datetime): The date of the revision.
            description (str): A description of the revision.
            source_code (io.BinaryIO): The report source code file, either a .py file or a .zip package.

        """
        # Create the revision metadata
        json: PayloadType = {
            'revision_date': revision_date.isoformat(),
            'description': description,
        }
        if project is not None:
            json['project'] = project
        r = self._client.request(f'{self._path}/revisions', 'PUT', json=json)
        revision = ReportSourceCodeRevision.model_validate_json(r.text)

        # Upload the source code
        r = self._client.request(
            f'{self._path}/revisions/{revision.code}', 'PATCH')
        upload_url = r.json()['target_source_code_url']
        upload_url = self._client.session().put(upload_url,
                                                data=source_code, verify=self._client.verify_ssl)
        if not upload_url.ok:
            raise MercutoHTTPException(
                f"Failed to upload report source code: {upload_url.status_code} {upload_url.text}",
                upload_url.status_code
            )
        return revision



from typing import Optional

import pytz

from ...modules.data import SecondaryDataSample
from .generic_csv import parse_generic_csv_file


def parse_campbell_file(filename: str, label_to_channel_code: dict[str, str],
                        timezone: Optional[pytz.BaseTzInfo] = None) -> list[SecondaryDataSample]:
    return parse_generic_csv_file(
        filename, label_to_channel_code, header_index=1, data_index=2, timezone=timezone)



import base64
import logging
from types import FunctionType
from typing import Any, Callable

logger = logging.getLogger(__name__)


class EnforceOverridesMeta(type):
    """
    Helper for patching mock implementations.

    Assign this as a metaclass when overriding a single base class, and all of the base class' methods
    must be overridden and cannot be called.

    Example usage:

    ```
    class BaseClass():
        def method1(self):
            pass
        def method2(self):
            pass

    class MockClass(BaseClass, metaclass=EnforceOverridesMeta):
        def method1(self):
            return "Mocked!"

    mock = MockClass()
    mock.method1() # Okay
    mock.method2() # Raises NotImplementedError
    ```

    To exclude specific methods from requiring overrides, add them to the `__exclude_enforce__` set:

    ```
    class MockClass(BaseClass, metaclass=EnforceOverridesMeta):
        __exclude_enforce__ = {BaseClass.method2}

        def method1(self):
            return "Mocked!"

    mock = MockClass()
    mock.method1() # Okay
    mock.method2() # Okay
    ```
    """
    def __init__(cls: type[Any], name: str, bases: tuple[type, ...], namespace: dict[str, Any]) -> None:
        type.__init__(cls, name, bases, namespace)

        # Assumes single inheritance from Base
        if len(bases) != 1:
            raise TypeError("EnforceOverridesMeta can only be used with a single base class.")

        base = bases[0]

        excluded: set[FunctionType] = getattr(cls, "__exclude_enforce__", set())

        for attr in dir(base):
            if attr.startswith('__'):
                continue

            base_method = getattr(base, attr)
            if isinstance(base_method, FunctionType) and base_method not in excluded and attr not in namespace:
                def make_error_method(attr_name: str) -> Callable[..., None]:
                    def error_method(self: Any, *args: Any, **kwargs: Any) -> None:
                        raise NotImplementedError(f"Mocker '{cls.__name__}' does not currently support method '{attr_name}'")
                    return error_method

                setattr(cls, attr, make_error_method(attr))


def create_data_url(mime_type: str, data: bytes) -> str:
    encoded_data = base64.b64encode(data).decode('utf-8')
    return f"data:{mime_type};base64,{encoded_data}"



import fnmatch
import logging
import os
from datetime import datetime
from typing import Optional

import pytz

from .. import MercutoClient, MercutoHTTPException
from ..modules.core import Project
from ..modules.data import (Channel, ChannelClassification, Datatable,
                            SecondaryDataSample)
from ..modules.media import Camera
from ..util import batched, get_my_public_ip
from .parsers import detect_parser

logger = logging.getLogger(__name__)

NON_RETRYABLE_ERRORS = {400, 404, 409}  # HTTP status codes that indicate non-retryable errors


def _get_file_mtime(file_path: str, increment: int) -> datetime:
    """
    Returns the file modification time rounded down to the nearest increment.
    """
    ts = os.path.getmtime(file_path)
    ts = ts - (ts % increment)
    return datetime.fromtimestamp(ts).astimezone()


class MercutoIngester:
    def __init__(self, project_code: str, api_key: str,
                 hostname: str = 'https://api.rockfieldcloud.com.au',
                 verify_ssl: bool = True,
                 timezone: Optional[str] = None,
                 camera_code: Optional[str] = None) -> None:
        """
        :param project_code: The Mercuto project code to ingest data into.
        :param api_key: The API key to use for authentication.
        :param hostname: The Mercuto server hostname.
        :param verify_ssl: Verify SSL certificates for the target server when using https. Default True.
        :param timezone: The timezone to use for data uploads as a string (e.g. 'Australia/Melbourne').
        :param camera_code: Optional camera code to associate with image uploads. If not provided, image uploads will error.
        """
        self._client = MercutoClient(url=hostname, verify_ssl=verify_ssl)
        self._api_key = api_key
        self._project_code = project_code
        self._timezone = timezone
        self._timezone_tzinfo = pytz.timezone(timezone) if timezone else None
        self._camera_code = camera_code

        self._project: Optional[Project] = None
        self._secondary_channels: Optional[list[Channel]] = None
        self._datatables: Optional[list[Datatable]] = None
        self._camera: Optional[Camera] = None

        self._channel_map: dict[str, str] = {}

    def _refresh_mercuto_data(self) -> None:
        with self._client.as_credentials(api_key=self._api_key) as client:
            self._project = client.core().get_project(self._project_code)
            assert self._project.code == self._project_code

            self._secondary_channels = client.data().list_channels(self._project_code, classification=ChannelClassification.SECONDARY)
            self._datatables = client.data().list_datatables(self._project_code)
            if self._camera_code is not None:
                self._camera = client.media().get_camera(self._camera_code)

        self._channel_map.update({c.label: c.code for c in self._secondary_channels})

    def _can_process(self) -> bool:
        return self._project is not None and self._secondary_channels is not None and self._datatables is not None

    def update_mapping(self, mapping: dict[str, str]) -> None:
        """
        Update the channel label to channel code mapping.
        """
        self._channel_map.update(mapping)
        logger.info(f"Updated channel mapping: {self._channel_map}")

    @property
    def project_code(self) -> str:
        return self._project_code

    def ping(self) -> None:
        """
        Ping the Mercuto serverto update the last seen IP address.
        """
        ip = get_my_public_ip()
        with self._client.as_credentials(api_key=self._api_key) as client:
            client.core().ping_project(self.project_code, ip_address=ip)
            logging.info(f"Pinged Mercuto server from IP: {ip} for project: {self.project_code}")

    def matching_datatable(self, filename: str) -> str | None:
        """
        Check if any datatables on the project match this file name.
        Returns the datatable code if a match is found, otherwise None.
        """
        if self._datatables is None:
            raise ValueError("Datatables not loaded. Call _refresh_mercuto_data() first.")

        basename = os.path.basename(filename)

        def matches(test: str) -> bool:
            """
            test should be a pattern or a filename.
            E.g. "my_data.csv" or "my_data*.csv", or "/path/to/my_data*.csv"
            Do wildcard matching as well as prefix matching.
            """
            test_base = os.path.basename(test)
            if fnmatch.fnmatch(basename, test_base):
                return True
            lhs, _ = os.path.splitext(test_base)
            if basename.startswith(lhs):
                return True
            return False

        for dt in self._datatables:
            # Match using datatable pattern
            if matches(dt.name):
                return dt.code
        return None

    def _upload_samples(self, samples: list[SecondaryDataSample]) -> bool:
        """
        Upload samples to the Mercuto project.
        """
        try:
            with self._client.as_credentials(api_key=self._api_key) as client:
                for batch in batched(samples, 500):
                    client.data().insert_secondary_samples(self.project_code, batch)
            return True
        except MercutoHTTPException as e:
            logger.error(f"Failed to upload samples: {e}")
            if e.status_code in NON_RETRYABLE_ERRORS:
                logger.exception(
                    "Error indicates bad file that should not be retried. Skipping.")
                return True
            else:
                return False

    def _upload_file(self, file_path: str, datatable_code: str) -> bool:
        """
        Upload a file to the Mercuto project.
        """
        logging.info(f"Uploading file {file_path} to datatable {datatable_code} in project {self.project_code}")
        try:
            with self._client.as_credentials(api_key=self._api_key) as client:
                client.data().upload_file(
                    project=self.project_code,
                    datatable=datatable_code,
                    file=file_path,
                    timezone=self._timezone
                )
            return True
        except MercutoHTTPException as e:
            logger.error(f"Failed to upload file {file_path} to datatable {datatable_code}: {e}")
            if e.status_code in NON_RETRYABLE_ERRORS:
                logger.exception(
                    "Error indicates bad file that should not be retried. Skipping.")
                return True
            else:
                return False

    def process_file(self, file_path: str) -> bool:
        """
        Process the received file.
        Returns True if processed successfully or should not be retried, False if processing failed and should be retried.
        """

        if not self._can_process():
            logging.info("Refreshing Mercuto data...")
            self._refresh_mercuto_data()
            if not self._can_process():
                logging.error("Failed to refresh Mercuto data. Cannot process file yet.")
                return False

        logging.info(f"Processing file: {file_path}")

        ext = os.path.splitext(file_path)[1]
        if ext in ['.dat', '.csv']:
            return self._process_data_file(file_path)
        elif ext in ['.jpg', '.jpeg', '.png', '.bmp', '.tiff']:
            return self._process_image_file(file_path)
        else:
            logger.error(f"Unsupported file extension: {ext} for file: {file_path}")
            # We mark unsupported files as processed to avoid retrying
            return True

    def _process_data_file(self, file_path: str) -> bool:
        """
        Process a data file specifically.
        Supported extensions: .dat, .csv
        """
        assert file_path.endswith(('.dat', '.csv'))
        datatable_code = self.matching_datatable(file_path)
        if datatable_code:
            logger.info(f"Matched datatable code: {datatable_code} for file: {file_path}")
            return self._upload_file(file_path, datatable_code)
        else:
            parser = detect_parser(file_path)
            samples = parser(file_path, self._channel_map, timezone=self._timezone_tzinfo)
            if not samples:
                logging.warning(f"No samples found in file: {file_path}")
                return True
            return self._upload_samples(samples)

    def _process_image_file(self, file_path: str) -> bool:
        """
        Process an image file specifically.
        For .jpg, .png, etc.
        """
        assert file_path.endswith(('.jpg', '.jpeg', '.png', '.bmp', '.tiff'))
        if self._camera is None:
            logger.error("No camera specified for image upload. Cannot process image file.")
            return False

        logging.info(f"Uploading image file {file_path} to camera {self._camera.code} in project {self.project_code}")
        timestamp = _get_file_mtime(file_path, increment=1)

        try:
            with self._client.as_credentials(api_key=self._api_key) as client:
                client.media().upload_image(
                    filename=file_path,
                    project=self.project_code,
                    camera=self._camera.code,
                    timestamp=timestamp,
                    event=None,
                    filedata=None
                )
            return True
        except MercutoHTTPException as e:
            logger.error(f"Failed to upload image file {file_path} to camera {self._camera.code}: {e}")
            if e.status_code in NON_RETRYABLE_ERRORS:
                logger.exception(
                    "Error indicates bad file that should not be retried. Skipping.")
                return True
            else:
                return False



import itertools
import shutil
from datetime import timedelta
from pathlib import Path
from typing import Iterable, Iterator, TypeVar

import requests


def timedelta_isoformat(td: timedelta) -> str:
    """
    ISO 8601 encoding for Python timedelta object.
    Taken from pydantic source:
        https://github.com/pydantic/pydantic/blob/3704eccce4661455acdda1cdcf716bd4b3382e08/pydantic/deprecated/json.py#L135-L140

    """
    minutes, seconds = divmod(td.seconds, 60)
    hours, minutes = divmod(minutes, 60)
    return f'{"-" if td.days < 0 else ""}P{abs(td.days)}DT{hours:d}H{minutes:d}M{seconds:d}.{td.microseconds:06d}S'


def get_my_public_ip() -> str:
    """
    Fetches the public IP address of the machine making the request.
    Uses the 'checkip.amazonaws.com' service to retrieve the IP address.
    :return The public IP address as a string in the form 'x.x.x.x'.
    :raises
        requests.RequestException: If the request to the IP service fails.
        requests.Timeout: If the request times out.
    """
    r = requests.get('https://checkip.amazonaws.com', timeout=30)
    r.raise_for_status()
    return r.content.decode().strip()


def get_directory_size(directory: str) -> int:
    """
    Returns the total size (in bytes) of the target directory, including all subdirectories.

    :param directory: Path to the target directory.
    :return: Total size in bytes.
    """
    dir_path = Path(directory)
    return sum(f.stat().st_size for f in dir_path.rglob("*") if f.is_file())


def get_free_space_excluding_files(directory: str) -> int:
    """
    Estimates the free bytes on the partition of the target directory
    as if the directory itself were empty.

    In other words, it adds back the space currently consumed by files
    within the target directory to the partition's free space.

    :param directory: Path to the target directory.
    :return: Estimated free bytes if the directory were empty.
    """
    # Get partition's free space (already excludes the directory's files)
    total, _, free = shutil.disk_usage(directory)

    # Calculate the total size of files in the directory
    files_size = get_directory_size(directory)

    # Add back the directory's file sizes to estimate free space if empty
    # Clamp to the partition's total capacity just in case
    return min(total, free + files_size)


T = TypeVar('T')

try:
    from itertools import batched  # type: ignore
except ImportError:
    # Python < 3.12
    def batched(iterable: Iterable[T], n: int) -> Iterator[tuple[T, ...]]:  # type: ignore[no-redef]
        """
        Implementation of itertools.batched for < Python 3.12
        """
        it = iter(iterable)
        while True:
            chunk = tuple(itertools.islice(it, n))
            if not chunk:
                break
            yield chunk



import os
import sqlite3
import tempfile
import time
from typing import Generator, Iterator, Tuple

import pytest

from ...ingester.processor import FileProcessor


def mock_process_callback(filepath: str) -> bool:
    return "success" in filepath  # Simulate success if filename contains "success"


@pytest.fixture
def buffer_directory() -> Iterator[str]:
    with tempfile.TemporaryDirectory() as temp_dir:
        yield temp_dir


@pytest.fixture
def work_directory() -> Iterator[str]:
    with tempfile.TemporaryDirectory() as temp_dir:
        yield temp_dir


@pytest.fixture
def database_path(work_directory: str) -> Iterator[str]:
    yield os.path.join(work_directory, "test_buffer.db")


@pytest.fixture
def temp_env(buffer_directory: str, database_path: str) -> Generator[Tuple[FileProcessor, str, str], None, None]:
    """Setup temporary directory and database"""

    processor: FileProcessor = FileProcessor(
        buffer_dir=buffer_directory,
        db_path=database_path,
        max_files=3,
        max_attempts=2,
        process_callback=mock_process_callback
    )

    yield processor, buffer_directory, database_path


def test_init_db(temp_env: Tuple[FileProcessor, str, str]) -> None:
    """Verify database initialization"""
    _, _, db_path = temp_env
    conn: sqlite3.Connection = sqlite3.connect(db_path)
    cursor: sqlite3.Cursor = conn.cursor()
    cursor.execute(
        "SELECT name FROM sqlite_master WHERE type='table' AND name='file_buffer'")
    assert cursor.fetchone() is not None  # Table should exist
    conn.close()


def test_register_file(temp_env: Tuple[FileProcessor, str, str]) -> None:
    """Check if new files are correctly registered in the database"""
    processor, buffer_dir, _ = temp_env
    test_file: str = os.path.join(buffer_dir, "file1.txt")

    # Create a test file
    with open(test_file, "w") as f:
        f.write("Test content")

    processor.add_file_to_db(test_file)

    conn: sqlite3.Connection = sqlite3.connect(processor.get_db_path())
    cursor: sqlite3.Cursor = conn.cursor()
    cursor.execute(
        "SELECT filename FROM file_buffer WHERE filename = 'file1.txt'")
    assert cursor.fetchone() is not None  # File should be registered
    conn.close()


def test_file_processing(temp_env: Tuple[FileProcessor, str, str]) -> None:
    """Verify file processing and persistence"""
    processor, buffer_dir, _ = temp_env
    test_file: str = os.path.join(buffer_dir, "success_file.txt")

    with open(test_file, "w") as f:
        f.write("Test content")

    processor.add_file_to_db(test_file)
    processor.process_next_file()

    conn: sqlite3.Connection = sqlite3.connect(processor.get_db_path())
    cursor: sqlite3.Cursor = conn.cursor()
    cursor.execute(
        "SELECT status FROM file_buffer WHERE filename = 'success_file.txt'")
    # File should be marked as processed
    assert cursor.fetchone()[0] == "processed"
    conn.close()


def test_file_processing_in_order(temp_env: Tuple[FileProcessor, str, str]) -> None:
    """Verify file processing and persistence"""
    processor, buffer_dir, _ = temp_env

    # Make some test files but with a reverse name so the order on the filesystem is not the same as the order in the database
    for i in reversed(range(5)):
        test_file: str = os.path.join(buffer_dir, f"success_file_{i}.txt")
        with open(test_file, "w") as f:
            f.write("Test content")

        processor.add_file_to_db(test_file)

    for i in reversed(range(5)):
        processed = processor.process_next_file()
        assert processed is not None, "Should process a file"
        name = os.path.basename(processed)
        assert name == f'success_file_{i}.txt', f"Should process files in order, got {name}"


def test_retry_attempts(temp_env: Tuple[FileProcessor, str, str]) -> None:
    """Ensure failed files are retried up to max_attempts"""
    processor, buffer_dir, _ = temp_env
    test_file: str = os.path.join(buffer_dir, "fail_file.txt")

    with open(test_file, "w") as f:
        f.write("Test content")

    processor.add_file_to_db(test_file)

    for _ in range(2):
        processor.process_next_file()

    conn: sqlite3.Connection = sqlite3.connect(processor.get_db_path())
    cursor: sqlite3.Cursor = conn.cursor()
    cursor.execute(
        "SELECT attempts FROM file_buffer WHERE filename = 'fail_file.txt'")
    # Should match max_attempts
    assert cursor.fetchone()[0] == 2
    conn.close()


@pytest.mark.parametrize("filename_part", ["success", "unprocessed", "failed"])
def test_cleanup_old_files_different_processing_status(temp_env: Tuple[FileProcessor, str, str], filename_part: str) -> None:
    """Ensure cleanup keeps only max_files, no matter the file status"""
    processor, buffer_dir, _ = temp_env

    # Create more files than max_files limit
    for i in range(5):
        test_file: str = os.path.join(buffer_dir, f"{filename_part}{i}.txt")
        with open(test_file, "w") as f:
            f.write("Test content")
        processor.add_file_to_db(test_file)
        if filename_part != 'unprocessed':
            # Process files that are not unprocessed
            processor.process_next_file()

    processor.cleanup_old_files()

    conn: sqlite3.Connection = sqlite3.connect(processor.get_db_path())
    cursor: sqlite3.Cursor = conn.cursor()
    cursor.execute(
        "SELECT COUNT(*) FROM file_buffer")
    # Should retain only max_files
    assert cursor.fetchone()[0] == 3

    # The last fiiles should be index 2,3,4
    cursor.execute(
        "SELECT filename FROM file_buffer ORDER BY timestamp ASC")
    remaining_files: list[tuple[str]] = cursor.fetchall()
    expected_files = [f"{filename_part}2.txt", f"{filename_part}3.txt", f"{filename_part}4.txt"]
    assert [f[0] for f in remaining_files] == expected_files

    conn.close()


def test_scan_existing_files(temp_env: Tuple[FileProcessor, str, str]) -> None:
    """Ensure that new files added are detected"""
    processor, buffer_dir, _ = temp_env

    for i in range(5):
        test_file = os.path.join(buffer_dir, f"file{i}.txt")
        with open(test_file, "w") as f:
            f.write("Test content")
        time.sleep(0.05)

    processor.scan_existing_files()
    conn: sqlite3.Connection = sqlite3.connect(processor.get_db_path())
    cursor: sqlite3.Cursor = conn.cursor()
    cursor.execute(
        "SELECT filename FROM file_buffer ORDER BY timestamp asc")

    files: list[tuple[str]] = cursor.fetchall()
    conn.close()
    assert files == [('file0.txt',), ('file1.txt',), ('file2.txt',), ('file3.txt',), ('file4.txt',)]


def test_scan_existing_files_that_havnt_been_processed(temp_env: Tuple[FileProcessor, str, str]) -> None:
    """Ensure that new files added are detected"""
    processor, buffer_dir, _ = temp_env

    for i in range(5):
        test_file = os.path.join(buffer_dir, f"file{i}.txt")
        with open(test_file, "w") as f:
            f.write("Test content")
        time.sleep(0.05)

        if i < 3:
            # Simulate adding for the first 3 files
            processor.add_file_to_db(test_file)
        if i < 2:
            # Simulate processing for the first 2 files
            processor.process_next_file()

    time.sleep(0.05)
    processor.scan_existing_files()
    conn: sqlite3.Connection = sqlite3.connect(processor.get_db_path())
    cursor: sqlite3.Cursor = conn.cursor()
    cursor.execute(
        "SELECT filename FROM file_buffer ORDER BY timestamp asc")

    files: list[tuple[str]] = cursor.fetchall()
    conn.close()
    assert files == [('file0.txt',), ('file1.txt',), ('file2.txt',), ('file3.txt',), ('file4.txt',)]


def test_ensure_free_space(database_path: str, buffer_directory: str) -> None:
    """Ensure that keeping X MB free space works correctly"""
    remaining_free_space_mb = 501

    def mock_free_space_checker(dir: str) -> float:
        """
        Mock free space checker that returns a controlled value.
        Returns remaining_free_space_mb - number of files in the buffer directory.
        """
        n_files = len(os.listdir(buffer_directory))
        return remaining_free_space_mb - n_files

    processor = FileProcessor(
        buffer_dir=buffer_directory,
        db_path=database_path,
        max_files=10,
        max_attempts=2,
        process_callback=mock_process_callback,
        target_free_space_mb=500,
        free_space_checker=mock_free_space_checker
    )

    def create_and_add_file(name: str) -> str:
        file_path = os.path.join(buffer_directory, name)
        with open(file_path, 'w') as f:
            f.write("Test content")
        processor.add_file_to_db(file_path)
        return file_path

    create_and_add_file('success_file_1.txt')
    create_and_add_file('success_file_2.txt')
    create_and_add_file('success_file_3.txt')
    create_and_add_file('success_file_4.txt')

    # There should be 4 files on disk
    assert len(os.listdir(buffer_directory)) == 4

    # Trigger cleanup, should not delete anything as we have enough free space
    remaining_free_space_mb = 504
    processor.cleanup_old_files()
    assert len(os.listdir(buffer_directory)) == 4

    # Reduce free space to trigger cleanup
    remaining_free_space_mb = 502
    processor.cleanup_old_files()
    # This should have deleted the oldest two files to maintain 500 MB free space
    # E.g. no deletes gives 498 MB free space, deleting one file gives 499 MB free space,
    # deleting two files gives 500 MB free space
    assert len(os.listdir(buffer_directory)) == 2
    remaining_files = os.listdir(buffer_directory)
    assert set(remaining_files) == {'success_file_3.txt', 'success_file_4.txt'}


def test_ensure_free_space_with_externally_modified_files(database_path: str, buffer_directory: str) -> None:
    """
    Ensure that keeping X MB free space works correctly under the following:
    - Files may be deleted from disk externally (not via the processor)
    """
    remaining_free_space_mb = 501

    def mock_free_space_checker(dir: str) -> float:
        """
        Mock free space checker that returns a controlled value.
        Returns remaining_free_space_mb - number of files in the buffer directory.
        """
        n_files = len(os.listdir(buffer_directory))
        return remaining_free_space_mb - n_files

    processor = FileProcessor(
        buffer_dir=buffer_directory,
        db_path=database_path,
        max_files=10,
        max_attempts=2,
        process_callback=mock_process_callback,
        target_free_space_mb=500,
        free_space_checker=mock_free_space_checker
    )

    def create_and_add_file(name: str) -> str:
        file_path = os.path.join(buffer_directory, name)
        with open(file_path, 'w') as f:
            f.write("Test content")
        processor.add_file_to_db(file_path)
        return file_path

    file1 = create_and_add_file('success_file_1.txt')
    file2 = create_and_add_file('success_file_2.txt')
    create_and_add_file('success_file_3.txt')
    create_and_add_file('success_file_4.txt')

    # There should be 4 files on disk
    assert len(os.listdir(buffer_directory)) == 4

    # Manually delete files 1 and 2 to simulate external deletion
    os.remove(file1)
    os.remove(file2)

    # Reduce free space to trigger cleanup
    remaining_free_space_mb = 100
    processor.cleanup_old_files()
    # This should still delete all files (even though 1 and 2 are already gone) to maintain 500 MB free space
    # The current free space of 100MB will never be reached, so all files should be deleted.
    assert len(os.listdir(buffer_directory)) == 0


def test_cleanup_max_files_with_externally_modified_files(database_path: str, buffer_directory: str) -> None:
    """
    Ensure that keeping X max files works correctly under the following:
    - Files may be deleted from disk externally (not via the processor)
    """

    processor = FileProcessor(
        buffer_dir=buffer_directory,
        db_path=database_path,
        max_files=1,
        max_attempts=2,
        process_callback=mock_process_callback,
        target_free_space_mb=None,
    )

    def create_and_add_file(name: str) -> str:
        file_path = os.path.join(buffer_directory, name)
        with open(file_path, 'w') as f:
            f.write("Test content")
        processor.add_file_to_db(file_path)
        return file_path

    file1 = create_and_add_file('success_file_1.txt')
    file2 = create_and_add_file('success_file_2.txt')
    create_and_add_file('success_file_3.txt')
    create_and_add_file('success_file_4.txt')

    # There should be 4 files on disk
    assert len(os.listdir(buffer_directory)) == 4

    # Manually delete files 1 and 2 to simulate external deletion
    os.remove(file1)
    os.remove(file2)

    # Trigger cleanup
    processor.cleanup_old_files()
    # THis should have deleted all but one file to maintain max_files=1
    assert len(os.listdir(buffer_directory)) == 1
    # It should be the last file added
    remaining_files = os.listdir(buffer_directory)
    assert set(remaining_files) == {'success_file_4.txt'}



from typing import Mapping, Optional


class IAuthenticationMethod:
    def update_header(self, header: dict[str, str]) -> None:
        return

    def unique_key(self) -> str:
        raise NotImplementedError(f"unique_key not implemented for type {self.__class__.__name__}")


class ApiKeyAuthentication(IAuthenticationMethod):
    def __init__(self, api_key: str) -> None:
        self.api_key = api_key

    def update_header(self, header: dict[str, str]) -> None:
        header['X-Api-Key'] = self.api_key

    def unique_key(self) -> str:
        return f'api-key:{self.api_key}'


class ServiceTokenAuthentication(IAuthenticationMethod):
    def __init__(self, service_token: str) -> None:
        self.service_token = service_token

    def update_header(self, header: dict[str, str]) -> None:
        header['X-Service-Token'] = self.service_token

    def unique_key(self) -> str:
        return f'service-token:{self.service_token}'


class AuthorizationHeaderAuthentication(IAuthenticationMethod):
    def __init__(self, authorization_header: str) -> None:
        if not authorization_header.startswith('Bearer '):
            authorization_header = 'Bearer ' + authorization_header
        self.authorization_header = authorization_header

    def update_header(self, header: dict[str, str]) -> None:
        header['Authorization'] = self.authorization_header

    def unique_key(self) -> str:
        return f'auth-bearer:{self.authorization_header}'


class NullAuthenticationMethod(IAuthenticationMethod):
    def update_header(self, header: dict[str, str]) -> None:
        pass

    def unique_key(self) -> str:
        return 'null-authentication'


def create_authentication_method(api_key: Optional[str] = None,
                                 service_token: Optional[str] = None,
                                 bearer_token: Optional[str] = None,
                                 headers: Optional[Mapping[str, str]] = None) -> IAuthenticationMethod:
    if api_key is not None and service_token is not None and headers is not None and bearer_token is not None:
        raise ValueError("Only one of api_key or service_token or bearer_token can be provided")
    if headers is not None:
        api_key = headers.get('X-Api-Key', None)
        service_token = headers.get('X-Service-Token', None)
        bearer_token = headers.get('Authorization', None)
    if api_key is not None:
        return ApiKeyAuthentication(api_key)
    elif service_token is not None:
        return ServiceTokenAuthentication(service_token)
    elif bearer_token is not None:
        return AuthorizationHeaderAuthentication(bearer_token)
    else:
        return NullAuthenticationMethod()



from .client import MercutoClient
from .exceptions import MercutoClientException, MercutoHTTPException

__all__ = ['MercutoClient', 'MercutoHTTPException', 'MercutoClientException']


def connect(*args, **kwargs) -> MercutoClient:
    return MercutoClient().connect(*args, **kwargs)



import logging
import shutil
import subprocess
import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path, PosixPath
from typing import Any, Callable, Optional
from urllib.parse import ParseResult, parse_qs, unquote
from urllib.request import url2pathname

logger = logging.getLogger(__name__)


class IBackupHandler(ABC):
    """
    Abstract base class for backup handlers.
    """

    def __init__(self, url: ParseResult):
        self.url = url
        self.validate_url()
        logger.debug(f'{self} backup location set to: {url.geturl()}')

    def __call__(self, filename: str) -> bool:
        """
        Provide a callable interface to process a file.
        :param filename: The file to process.
        """
        try:
            result = self.process_file(filename)
            if not result:
                logger.error(f"{self} Failed to process {filename}")
            else:
                logger.debug(f"{self} processed {filename}")
            return result
        except Exception:
            return False

    def decode_query(self) -> dict[str, Any]:
        """
        Decodes the query parameters from the URL.
        :return: A dictionary of query parameters (e.g. http://example.com?param1=value1 ==> {'param1': 'value1'})
        """
        return parse_qs(unquote(self.url.query))

    def validate_url(self):
        """
        Validates the URL components.
         Raises an exception if the URL is invalid.

        Should be implemented by subclasses
        """
        pass

    @abstractmethod
    def process_file(self, filename: str) -> bool:
        """
        Processes the given file.
        Should be implemented by subclasses.
        """
        raise NotImplementedError()


class CSCPBackup(IBackupHandler):
    """
    Compressed SCP Backup handler (cscp://username@hostname:port/path?private_key=xxx).
    Uses SCP to copy files to a remote server, with optional post-transfer script execution.
    Use query parameters to specify additional options:
    - private_key: Path to the private key file for authentication.
    - script: A script to run on the remote server after file transfer. The script can use
      the placeholder '{destination}' to refer to the path of the transferred file.
    - disable_strict_checking: If set to 'yes', disables strict checking of host keys, known_hosts file is not used,
        and user private keys can be open.
        It adds the following options to the scp/ssh command: oStrictHostKeyChecking=no, oUserKnownHostsFile=/dev/null
    """
    @dataclass
    class SCPBackupParams:
        private_key: Optional[Path] = None
        script: Optional[str] = None
        disable_strict_checking: Optional[str] = None

        @staticmethod
        def load(config: dict[str, Any]) -> 'CSCPBackup.SCPBackupParams':
            """
            Loads SCPBackupParams from a configuration dictionary.
            :param config: Configuration dictionary.
            :return: An instance of CSCPBackup.SCPBackupParams.
            """
            return CSCPBackup.SCPBackupParams(**config)

        @staticmethod
        def load_qs(query: dict[str, Any]) -> 'CSCPBackup.SCPBackupParams':
            """
            Load SCPBackupParams from a query string dictionary.
            :param query: Query string dictionary.
            """
            result: dict[str, Any] = {}
            keys = list(query.keys())
            known_keys = ['private_key', 'script', 'disable_strict_checking']
            for _key in known_keys:
                if _key in query:
                    if len(query[_key]) > 1:
                        raise RuntimeError(f"Multiple {_key} entries found")
                    result[_key] = query[_key][0]
                    keys.remove(_key)
            if len(keys) > 0:
                raise RuntimeError(f"Unknown key {keys}")

            return CSCPBackup.SCPBackupParams.load(result)

    def __init__(self, url: ParseResult):
        self.params: Optional[CSCPBackup.SCPBackupParams] = None
        super().__init__(url)

        self.port = 22
        if self.url.port is not None:
            self.port = self.url.port

        self.backup_path = Path(url2pathname(self.url.path))

    def validate_url(self):
        if self.url.hostname is None:
            raise RuntimeError("No hostname specified for backup")
        query = self.decode_query()
        self.params = CSCPBackup.SCPBackupParams.load_qs(query)

    def process_file(self, filename: str) -> bool:
        if self.send_file(filename):
            return self.run_script(filename)
        return False

    def send_file(self, filename: str) -> bool:
        if self.params is None:
            raise ValueError("Backup parameters not set")
        command = ['scp', '-O', '-oBatchMode=yes']
        if self.params.disable_strict_checking == 'yes':
            command.append('-oStrictHostKeyChecking=no')
            command.append('-oUserKnownHostsFile=/dev/null')

        dest = ""
        if self.url.username is not None:
            dest = f"{self.url.username}@"

        dest = f'{dest}{self.url.hostname}'
        command.append('-P')
        command.append(str(self.port))

        if self.params.private_key is not None:
            command.append('-i')
            command.append(str(self.params.private_key))

        dest = f'{dest}:{self.backup_path}'

        command.append(filename)
        command.append(dest)

        try:
            logger.debug(f'Copy Command: {" ".join(command)}')
            result = subprocess.run(command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            logger.debug(f"STDOUT: {result.stdout.decode('utf-8', errors='ignore')}")
            logger.debug(f"STDERR: {result.stderr.decode('utf-8', errors='ignore')}")
            return result.returncode == 0

        except subprocess.CalledProcessError as e:
            print(e.stdout.decode("utf-8"), file=sys.stderr)
            print(e.stderr.decode("utf-8"), file=sys.stderr)
            return False

    def run_script(self, filename: str) -> bool:
        if self.params is None:
            return True
        elif self.params.script is None:
            return True
        command = ['ssh', '-oBatchMode=yes']

        if self.params.disable_strict_checking == 'yes':
            command.append('-oStrictHostKeyChecking=no')
            command.append('-oUserKnownHostsFile=/dev/null')
        if self.url.username is not None:
            command.append('-l')
            command.append(self.url.username)

        port = 22
        if self.url.port is not None:
            port = self.url.port

        command.append('-p')
        command.append(str(port))

        if self.params.private_key is not None:
            command.append('-i')
            command.append(str(self.params.private_key))

        if self.url.hostname is None:
            raise RuntimeError("No hostname specified for backup")
        command.append(self.url.hostname)

        dest_folder = PosixPath(self.backup_path)
        fn = Path(filename)

        command.append(self.params.script.format(destination=dest_folder / fn.name))
        logger.debug(f'Script Command: {" ".join(command)}')
        try:
            logger.debug(f'Script Command: {" ".join(command)}')
            result = subprocess.run(command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            logger.debug(f"STDOUT: {result.stdout.decode('utf-8', errors='ignore')}")
            logger.debug(f"STDERR: {result.stderr.decode('utf-8', errors='ignore')}")
            return result.returncode == 0

        except subprocess.CalledProcessError as e:
            print(e.stdout.decode("utf-8"), file=sys.stderr)
            print(e.stderr.decode("utf-8"), file=sys.stderr)
            return False


class FileBackup(IBackupHandler):
    """
    File Backup handler (file://)
    Copies files to a specified local directory.
    """

    def __init__(self, url: ParseResult):
        super().__init__(url)
        self.backup_path = Path(url2pathname(self.url.path))
        query = self.decode_query()
        self.create = False
        if 'create' in query:
            create = query['create']
            if not isinstance(create, list) or len(create) != 1 or not isinstance(create[0], str):
                raise ValueError(f"{self} create query element has wrong length: {len(create)}, expected 1")
            create = create[0].lower()
            if create in ['true', 'yes', 'y']:
                self.create = True

        if not self.create and not self.backup_path.exists():
            raise ValueError(f"{self.backup_path} backup path does not exist")

    def validate_url(self):
        if self.url.scheme.lower() != 'file':
            raise ValueError(f"{self} url scheme must be 'file'")

    def process_file(self, filename: str) -> bool:
        if not self.create and not self.backup_path.exists():
            raise ValueError(f"{self.backup_path} backup path does not exist")
        elif self.create and not self.backup_path.exists():
            self.backup_path.mkdir(parents=True, exist_ok=True)
            logger.info(f"Created backup path: {self.backup_path}")

        if not self.backup_path.is_dir():
            raise ValueError(f"{self.backup_path} backup path must be a directory")

        dest = self.backup_path / Path(filename).name
        shutil.copyfile(filename, dest)
        return True


class HTTPBackup(IBackupHandler):
    """
    HTTP Backup handler (http:// or https://).
    Uses HTTP POST to send files to a specified URL.
    """

    def validate_url(self):
        if self.url.scheme.lower() not in ['http', 'https']:
            raise ValueError(f"{self} url scheme must be 'http' or 'https'")

    def process_file(self, filename: str) -> bool:
        import requests  # import here to avoid dependency if not used
        with open(filename, "rb") as f:
            files = {"file": (Path(filename).name, f)}
            response = requests.post(self.url.geturl(), files=files)
            return response.ok


class FTPBackup(IBackupHandler):
    """
    FTP Backup handler (ftp:// or ftps://).
    Uses FTP to send files to a specified URL.
    """

    def __init__(self, url: ParseResult):
        super().__init__(url)
        import ftplib

        self._ftp: Optional[ftplib.FTP | ftplib.FTP_TLS] = None
        self._port = self.url.port or (21 if self.url.scheme.lower() == 'ftp' else 990)

        if self.url.scheme.lower() == 'ftps':
            self._ftp = ftplib.FTP_TLS()
        else:
            self._ftp = ftplib.FTP()

    def validate_url(self):
        if self.url.scheme.lower() not in ['ftp', 'ftps']:
            raise ValueError(f"{self} url scheme must be 'ftp' or 'ftps'")

        if self.url.hostname is None:
            raise ValueError(f"{self} url must specify a hostname")

    def process_file(self, filename: str) -> bool:
        # Send file via FTP
        try:
            assert self._ftp is not None
            assert self.url.hostname is not None
            self._ftp.connect(self.url.hostname, self._port)

            username = self.url.username or 'anonymous'
            password = self.url.password or ''
            self._ftp.login(username, password)

            dest_path = self.url.path
            with open(filename, 'rb') as f:
                self._ftp.storbinary(f'STOR {dest_path}/{Path(filename).name}', f)
            self._ftp.quit()
            return True

        except Exception as e:
            logger.error(f"FTP error: {e}")
            return False


def get_backup_handler(url: ParseResult) -> Callable[[str], bool]:
    match url.scheme.lower():
        case "file":
            return FileBackup(url)
        case "http":
            return HTTPBackup(url)
        case "https":
            return HTTPBackup(url)
        case "ftp":
            return FTPBackup(url)
        case "ftps":
            return FTPBackup(url)
        case "cscp":
            return CSCPBackup(url)
        case "scp":
            return CSCPBackup(url)
        case _:
            raise ValueError(f"Unsupported scheme: {url.scheme}")




import os
import shutil
from pathlib import Path

from .. import util


def test_get_free_space_excluding_files_with_real_dir(tmp_path: Path):
    # Start with an empty directory
    du = shutil.disk_usage(tmp_path)
    initial_estimate = util.get_free_space_excluding_files(str(tmp_path))

    # For an empty dir, estimate should match current free space
    assert initial_estimate == du.free

    # Create some files totaling a few MiB
    sizes = [1 * 1024 * 1024, 10 * 1024 * 1024, 512 * 1024]  # 1MiB, 10MiB, 512KiB
    for idx, size in enumerate(sizes, start=1):
        p = tmp_path / f"file_{idx}.bin"
        with open(p, "wb") as f:
            f.write(os.urandom(size))

    # Add a subdirectory with more files
    subdir = tmp_path / "subdir"
    subdir.mkdir()
    sub_sizes = [2 * 1024 * 1024, 3 * 1024 * 1024]  # 2MiB, 3MiB
    for idx, size in enumerate(sub_sizes, start=1):
        p = subdir / f"subfile_{idx}.bin"
        with open(p, "wb") as f:
            f.write(os.urandom(size))

    # After adding files, the function adds back their sizes, so the estimate
    # should remain effectively the same as before. Allow minor FS overhead.
    after_estimate = util.get_free_space_excluding_files(str(tmp_path))

    # Tolerance accounts for filesystem metadata/cluster rounding
    tolerance = 64 * 1024  # 64 KiB
    assert abs(after_estimate - initial_estimate) <= tolerance

    # Ensure the total size using the get_directory_size() function matches
    assert util.get_directory_size(str(tmp_path)) == sum(sizes) + sum(sub_sizes)



import requests

from ..exceptions import MercutoClientException, MercutoHTTPException

_PayloadValueType = str | float | int | None | bool
_PayloadListType = list[str] | list[float] | list[int] | list[_PayloadValueType]
_PayloadDictType = dict[str, '_PayloadValueType | _PayloadListType | _PayloadDictType']
PayloadType = dict[str, _PayloadValueType | _PayloadListType | _PayloadDictType]
_PayloadType = PayloadType  # For backwards compatibility


def raise_for_response(r: requests.Response) -> None:
    if 500 <= r.status_code < 600:
        raise MercutoClientException(f"Server error: {r.text}")
    if not (200 <= r.status_code < 300):
        try:
            detail = r.text
        except Exception:
            detail = str(r)
        raise MercutoHTTPException(detail, r.status_code)


_raise_for_response = raise_for_response  # For backwards compatibility



from datetime import timedelta
from typing import TYPE_CHECKING, Any, Literal, Optional

from pydantic import AwareDatetime, TypeAdapter

from .data import AggregationMethod

if TYPE_CHECKING:
    from ..client import MercutoClient

from . import PayloadType
from ._util import BaseModel


class Healthcheck(BaseModel):
    status: str


class Condition(BaseModel):
    class ConditionAggregationOption(BaseModel):
        method: AggregationMethod
        interval: timedelta
    code: str
    project: str
    description: Optional[str]
    check_type: Literal['event', 'interval']

    match_on: Optional[list[str]] = None
    match_type: Optional[Literal['channel-label-pattern', 'channel-code', 'device-label-pattern',
                                 'device-type-code', 'camera-label-pattern', 'camera-code']] = None
    field: Optional[str] = None
    upper_start_threshold: Optional[float] = None
    lower_start_threshold: Optional[float] = None
    upper_end_threshold: Optional[float] = None
    lower_end_threshold: Optional[float] = None
    aggregation: Optional[ConditionAggregationOption] = None
    duration_threshold: Optional[timedelta] = None
    type: Literal['channel-range', 'channel-offline', 'camera-offline']


class AlarmConditionGroup(BaseModel):
    same_target: Literal['none', 'exact', 'effective']
    conditions: list[Condition]


class Alarm(BaseModel):
    code: str
    project: str
    label: str
    severity: int
    contact_group: Optional[str]
    check_type: Literal['event', 'interval']
    condition_groups: list[AlarmConditionGroup]
    minimum_unique_targets: int


class ConditionLog(BaseModel):
    code: str
    condition: str
    created_at: AwareDatetime
    start_time: AwareDatetime
    end_time: AwareDatetime

    target_code: str
    target_type: Literal['channel', 'camera']
    start_value: Optional[float]
    end_value: Optional[float]
    peak_value: Optional[float]
    peak_time: Optional[AwareDatetime]
    event: Optional[str]
    is_still_active: bool


class AlarmLog(BaseModel):
    code: str
    fired_at: AwareDatetime
    active: bool
    acknowledged_at: Optional[AwareDatetime]
    event: Optional[str]
    severity: int
    alarm: str
    condition_logs: list[ConditionLog]


# --- TypeAdapters for lists ---
_ConditionListAdapter = TypeAdapter(list[Condition])
_AlarmListAdapter = TypeAdapter(list[Alarm])
_ConditionLogListAdapter = TypeAdapter(list[ConditionLog])
_AlarmLogListAdapter = TypeAdapter(list[AlarmLog])


class MercutoAlertService:
    def __init__(self, client: 'MercutoClient', path: str = '/v2/alerts') -> None:
        self._client = client
        self._path = path

    def healthcheck(self) -> Healthcheck:
        r = self._client.request(f"{self._path}/healthcheck", "GET")
        return Healthcheck.model_validate_json(r.text)

    # --- Conditions ---
    def list_conditions(self, project: str) -> list[Condition]:
        r = self._client.request(
            f"{self._path}/conditions", "GET", params={"project": project})
        return _ConditionListAdapter.validate_json(r.text)

    def create_condition(self, /, **args: Any) -> Condition:
        r = self._client.request(f"{self._path}/conditions", "POST", json=args)
        return Condition.model_validate_json(r.text)

    def get_condition(self, condition_code: str) -> Condition:
        r = self._client.request(
            f"{self._path}/conditions/{condition_code}", "GET")
        return Condition.model_validate_json(r.text)

    def update_condition(self, condition_code: str, /, **args: Any) -> Condition:
        r = self._client.request(
            f"{self._path}/conditions/{condition_code}", "PUT", json=args)
        return Condition.model_validate_json(r.text)

    def delete_condition(self, condition_code: str) -> None:
        self._client.request(
            f"{self._path}/conditions/{condition_code}", "DELETE")
        return None

    # --- Condition Logs ---
    def list_condition_logs(self, project: str) -> list[ConditionLog]:
        r = self._client.request(
            f"{self._path}/condition-logs", "GET", params={"project": project})
        return _ConditionLogListAdapter.validate_json(r.text)

    # --- Alarms ---
    def list_alarms(self, project: str) -> list[Alarm]:
        r = self._client.request(
            f"{self._path}/alarms", "GET", params={"project": project})
        return _AlarmListAdapter.validate_json(r.text)

    def create_alarm(self, /, **args: Any) -> Alarm:
        r = self._client.request(f"{self._path}/alarms", "POST", json=args)
        return Alarm.model_validate_json(r.text)

    def get_alarm(self, alarm_code: str) -> Alarm:
        r = self._client.request(f"{self._path}/alarms/{alarm_code}", "GET")
        return Alarm.model_validate_json(r.text)

    def update_alarm(self, alarm_code: str, /, **args: Any) -> Alarm:
        r = self._client.request(
            f"{self._path}/alarms/{alarm_code}", "PUT", json=args)
        return Alarm.model_validate_json(r.text)

    def delete_alarm(self, alarm_code: str) -> None:
        self._client.request(f"{self._path}/alarms/{alarm_code}", "DELETE")
        return None

    # --- Alarm Logs ---
    def list_alarm_logs(self, project: str,
                        alarm: Optional[str] = None,
                        acknowledged: Literal['yes', 'no', 'any'] = 'any') -> list[AlarmLog]:
        params: PayloadType = {"project": project,
                               "acknowledged": acknowledged}
        if alarm is not None:
            params["alarm"] = alarm
        r = self._client.request(
            f"{self._path}/alarm-logs", "GET", params=params)
        return _AlarmLogListAdapter.validate_json(r.text)

    def get_alarm_log(self, alarm_log_code: str) -> AlarmLog:
        r = self._client.request(
            f"{self._path}/alarm-logs/{alarm_log_code}", "GET")
        return AlarmLog.model_validate_json(r.text)

    def acknowledge_alarm_log(self, alarm_log_code: str) -> AlarmLog:
        r = self._client.request(
            f"{self._path}/alarm-logs/{alarm_log_code}/acknowledge", "POST")
        return AlarmLog.model_validate_json(r.text)



import logging
import uuid
from datetime import datetime
from datetime import timezone as py_timezone
from typing import Optional

from ..client import MercutoClient
from ..exceptions import MercutoHTTPException
from ..modules.core import (Device, DeviceChannel, DeviceType, Event, ItemCode,
                            MercutoCoreService, Project, ProjectStatus)
from ._utility import EnforceOverridesMeta

logger = logging.getLogger(__name__)


class MockMercutoCoreService(MercutoCoreService, metaclass=EnforceOverridesMeta):
    def __init__(self, client: 'MercutoClient'):
        super().__init__(client=client)
        self._events: dict[str, Event] = {}
        self._projects: dict[str, Project] = {}
        self._devices: dict[str, Device] = {}

    def ping_project(self, project: str, ip_address: str) -> None:
        if project not in self._projects:
            raise MercutoHTTPException(status_code=404, message=f"Project {project} not found")

        self._projects[project].status.last_ping = datetime.now(py_timezone.utc).isoformat()
        self._projects[project].status.ip_address = ip_address

    def get_project(self, code: str) -> Project:
        if code not in self._projects:
            raise MercutoHTTPException(status_code=404, message=f"Project {code} not found")
        return self._projects[code]

    def create_project(self, name: str, project_number: str, description: str, tenant: str,
                       timezone: str, latitude: Optional[float] = None,
                       longitude: Optional[float] = None) -> Project:
        project = Project(
            code=str(uuid.uuid4()),
            name=name,
            project_number=project_number,
            active=True,
            description=description,
            tenant=tenant,
            timezone=timezone,
            latitude=latitude,
            longitude=longitude,
            display_timezone=timezone,
            status=ProjectStatus(last_ping=None, ip_address=None),
            commission_date=datetime(2020, 1, 1, tzinfo=py_timezone.utc)
        )
        self._projects[project.code] = project
        return project

    def create_event(self, project: str, start_time: datetime, end_time: datetime) -> Event:
        event = Event(code=str(uuid.uuid4()), project=ItemCode(code=project), start_time=start_time, end_time=end_time, objects=[], tags=[])
        self._events[event.code] = event
        return event

    def get_event(self, event: str) -> Event:
        if event not in self._events:
            raise MercutoHTTPException(status_code=404, message=f"Event {event} not found")
        return self._events[event]

    def list_events(self, project: str,
                    start_time: Optional[datetime] = None,
                    end_time: Optional[datetime] = None,
                    limit: Optional[int] = None, offset: Optional[int] = 0,
                    ascending: bool = True) -> list[Event]:
        filtered = [event for event in self._events.values() if event.project.code == project]
        if start_time is not None:
            filtered = [event for event in filtered if event.start_time >= start_time]
        if end_time is not None:
            filtered = [event for event in filtered if event.end_time <= end_time]
        filtered.sort(key=lambda e: e.start_time, reverse=not ascending)
        if offset is not None:
            filtered = filtered[offset:]
        if limit is not None:
            filtered = filtered[:limit]
        return filtered

    def list_devices(self, project_code: str, limit: int, offset: int) -> list[Device]:
        return [device for device in self._devices.values() if device.project.code == project_code][offset:offset+limit]

    def get_device(self, device_code: str) -> Device:
        if device_code not in self._devices:
            raise MercutoHTTPException(status_code=404, message=f"Device {device_code} not found")
        return self._devices[device_code]

    def create_device(self,
                      project_code: str,
                      label: str,
                      device_type_code: str,
                      groups: list[str],
                      location_description: Optional[str] = None,
                      channels: Optional[list[DeviceChannel]] = None,
                      latitude: Optional[float] = None,
                      longitude: Optional[float] = None,
                      altitude: Optional[float] = None
                      ) -> Device:
        device = Device(
            code=str(uuid.uuid4()),
            project=ItemCode(code=project_code),
            label=label,
            device_type=DeviceType(
                code=device_type_code,
                description="A mock device type for testing",
                manufacturer="Mock Manufacturer",
                model_number="Model X",

            ),
            groups=groups,
            location_description=location_description,
            channels=channels or [],
        )
        self._devices[device.code] = device
        return device



import argparse
import atexit
import logging
import sys
from pathlib import Path

from zc.lockfile import LockError, LockFile  # type: ignore[import-untyped]

logger = logging.getLogger(__name__)


class PidFile:
    """
    Context manager for creating and removing a PID lock file.
    :param lock_file: Path to the lock file to create. If None, no lock file is created.
    :param content_template: Template for the content of the lock file. Default is '{pid}'.
    """

    def __init__(self, lock_file: Path | None = None, content_template: str = '{pid}'):
        self.content_template = content_template
        self.lock_path = Path(lock_file) if lock_file else None
        self.lock: LockFile | None = None

    def __cleanup(self):
        if self.lock_path is not None:
            self.lock_path.unlink(missing_ok=True)

    def __enter__(self):
        if self.lock_path is not None:
            self.lock = LockFile(self.lock_path, content_template=self.content_template)
            logger.warning(f'Created lock file {self.lock}')
            atexit.register(self.__cleanup)
        return self

    def __exit__(self, exc_type: type | None, exc_val: BaseException | None, exc_tb: object | None) -> None:
        if self.lock is not None:
            self.lock.close()
            self.lock = None
            self.__cleanup()
            atexit.unregister(self.__cleanup)


def main():
    locked = 0
    unlocked = 1
    parser = argparse.ArgumentParser()
    parser.add_argument('pidfile', type=Path)

    args = parser.parse_args()
    if args.pidfile.exists():
        try:
            lock = LockFile(args.pidfile)
            lock.close()
            print(f"pid file '{args.pidfile}' is not locked", file=sys.stderr)

        except LockError:
            print(f"pid file '{args.pidfile}' is locked", file=sys.stderr)
            sys.exit(locked)

    else:
        print(f"pid file '{args.pidfile}' does not exist", file=sys.stderr)
    sys.exit(unlocked)


if __name__ == '__main__':
    main()



from typing import Iterator

import pytest

from ... import MercutoClient
from ...mocks import mock_mercuto


@pytest.fixture
def client() -> Iterator[MercutoClient]:
    with mock_mercuto():
        client = MercutoClient("https://testserver")
        yield client



import json
from typing import Any


class MercutoClientException(Exception):
    pass


class MercutoHTTPException(MercutoClientException):
    def __init__(self, message: str, status_code: int) -> None:
        super().__init__(message)
        self.status_code = status_code
        self.message = message

    def json(self) -> Any:
        return json.loads(self.message)

    def __str__(self) -> str:
        return f"MercutoHTTPException(status_code='{self.status_code}', message='{self.message}')"



import base64
import io
import logging
import uuid
from datetime import datetime, timedelta, timezone
from typing import BinaryIO, Collection, Optional, TextIO

import pandas as pd

from ..client import MercutoClient
from ..exceptions import MercutoHTTPException
from ..modules.data import (AggregationMethod, AggregationOptions, Channel,
                            ChannelClassification, ChannelFormat, Datatable,
                            DatatableColumn, FileFormat, FrameFormat,
                            GetStatusRequestResponse, LatestDataSample,
                            MercutoDataService, MetricDataSample,
                            SecondaryDataSample, Units)
from ._utility import EnforceOverridesMeta

logger = logging.getLogger(__name__)


class MockMercutoDataService(MercutoDataService, metaclass=EnforceOverridesMeta):
    __exclude_enforce__ = {MercutoDataService.load_presigned_url,
                           MercutoDataService.load_metric_sample,
                           MercutoDataService.load_data_request}

    def __init__(self, client: 'MercutoClient'):
        super().__init__(client=client, path='/mock-data-service-method-not-implemented')
        self._secondary_and_primary_buffer = pd.DataFrame(columns=['channel', 'timestamp', 'value']).set_index(['channel', 'timestamp'])
        self._metric_buffer = pd.DataFrame(columns=['channel', 'timestamp', 'value', 'event']).set_index(['channel', 'timestamp'])

        self._known_requests: dict[str, GetStatusRequestResponse] = {}

        self._channels: dict[str, Channel] = {}
        self._datatables: dict[str, Datatable] = {}
        self._units: dict[str, Units] = {}

    def _last_timestamp(self, channel: str) -> Optional[datetime]:
        if channel in self._secondary_and_primary_buffer.index.get_level_values('channel'):
            return self._secondary_and_primary_buffer.loc[channel, :].index.get_level_values('timestamp').max()
        if channel in self._metric_buffer.index.get_level_values('channel'):
            return self._metric_buffer.loc[channel, :].index.get_level_values('timestamp').max()
        return None

    def _update_last_valid_samples(self) -> None:
        for channel in self._channels.values():
            channel.last_valid_timestamp = self._last_timestamp(channel.code)

    def list_channels(self,
                      project: str,
                      classification: Optional[ChannelClassification] = None,
                      aggregate: Optional[str] = None, metric: Optional[str] = None,
                      show_hidden: bool = False) -> list[Channel]:
        selection = filter(lambda ch: ch.project == project, self._channels.values())
        if classification is not None:
            selection = filter(lambda ch: ch.classification == classification, selection)
        if aggregate is not None:
            selection = filter(lambda ch: ch.aggregate == aggregate, selection)
        if metric is not None:
            selection = filter(lambda ch: ch.metric == metric, selection)
        return list(selection)

    def get_channel(self, code: str) -> Optional[Channel]:
        return self._channels.get(code)

    def create_channel(self, project: str,
                       label: str,
                       classification: ChannelClassification = ChannelClassification.SECONDARY,
                       sampling_period: Optional[timedelta] = None,
                       multiplier: float = 1.0, offset: float = 0.0,
                       value_range_min: Optional[float] = None, value_range_max: Optional[float] = None,
                       delta_max: Optional[float] = None,
                       units: Optional[str] = None,
                       aggregate: Optional[str] = None,
                       source: Optional[str] = None,
                       metric: Optional[str] = None) -> Channel:

        if multiplier != 1.0:
            logger.warning("MockMercutoDataService does not support channel multiplier.")
        if offset != 0.0:
            logger.warning("MockMercutoDataService does not support channel offset.")
        if value_range_min is not None:
            logger.warning("MockMercutoDataService does not support channel value_range_min.")
        if value_range_max is not None:
            logger.warning("MockMercutoDataService does not support channel value_range_max.")
        if delta_max is not None:
            logger.warning("MockMercutoDataService does not support channel delta_max.")
        if units is not None:
            logger.warning("MockMercutoDataService does not support channel units.")

        # Generate a unique UUID code
        code = str(uuid.uuid4())
        channel = Channel(code=code,
                          project=project,
                          units=None,
                          sampling_period=sampling_period,
                          classification=classification,
                          label=label,
                          metric=metric,
                          source=source,
                          aggregate=aggregate,
                          value_range_min=None,
                          value_range_max=None,
                          multiplier=multiplier,
                          offset=offset,
                          last_valid_timestamp=None,
                          is_wallclock_interval=False)
        self._channels[code] = channel

        return channel

    def update_channel(self, code: str, label: Optional[str] = None, units: Optional[str] = None,
                       metric: Optional[str] = None, multiplier: Optional[float] = None,
                       offset: Optional[float] = None) -> Channel:
        if code not in self._channels:
            raise MercutoHTTPException(status_code=404, message="Channel not found")
        channel = self._channels[code]
        if label is not None:
            channel.label = label
        if units is not None:
            channel.units = self.get_unit(units)
        if metric is not None:
            channel.metric = metric
        if multiplier is not None:
            channel.multiplier = multiplier
        if offset is not None:
            channel.offset = offset
        self._channels[code] = channel
        return channel

    def create_request(self,
                       start_time: datetime,
                       end_time: datetime,
                       project: Optional[str] = None,
                       channels: Optional[Collection[str]] = None,
                       classification: Optional[ChannelClassification] = None,
                       frame_format: FrameFormat = FrameFormat.SAMPLES,
                       file_format: FileFormat = FileFormat.PARQUET,
                       channel_format: ChannelFormat = ChannelFormat.CODE,
                       aggregation: Optional[AggregationOptions] = None,
                       timeout: float = 0) -> GetStatusRequestResponse:

        if channel_format != ChannelFormat.CODE:
            raise NotImplementedError(f"Unsupported channel format: {channel_format}")

        if channels is None and project is None:
            raise ValueError("Must supply either channels or project.")

        if channels is None and classification is None:
            raise ValueError("Must supply either channels or classification.")

        if channels is None:
            assert classification is not None
            assert project is not None
            channels = [ch.code for ch in self._channels.values() if ch.classification == classification and ch.project == project]

        assert channels is not None

        def load_from_buffer(buffer: pd.DataFrame) -> pd.DataFrame:
            # Filter by channels if provided
            if channels is not None:
                buffer = buffer[buffer.index.get_level_values('channel').isin(channels)]

            # Filter by time range
            buffer = buffer[
                (buffer.index.get_level_values('timestamp') >= start_time) &
                (buffer.index.get_level_values('timestamp') <= end_time)
            ]
            return buffer

        secondary_part = load_from_buffer(self._secondary_and_primary_buffer)
        metric_part = load_from_buffer(self._metric_buffer)[['value']]
        ts = pd.concat([secondary_part, metric_part], axis=0).sort_index()

        assert ts.columns == ['value']
        assert ts.index.names == ['channel', 'timestamp']

        # Apply aggregation options
        if aggregation is not None:
            if aggregation.method != AggregationMethod.MEAN:
                raise NotImplementedError(f"Unsupported aggregation method for MockClient: {aggregation.method}")
            if aggregation.interval != 'day' and aggregation.interval != timedelta(days=1):
                raise NotImplementedError(f"Unsupported aggregation interval for MockClient: {aggregation.interval}")
            if not aggregation.rolling:
                raise NotImplementedError("MockClient only supports rolling aggregation.")
            ts = (
                ts.reset_index()  # type: ignore
                .sort_values(['channel', 'timestamp'])
                .groupby('channel')
                .rolling('1D', on='timestamp', min_periods=1)['value']
                .mean()
                .reset_index()
                .set_index(['channel', 'timestamp'])
                .sort_index()
            )
            assert ts.columns == ['value']
            assert ts.index.names == ['channel', 'timestamp']
        if frame_format == FrameFormat.COLUMNS:
            ts = ts.reset_index(drop=False).pivot(index='timestamp',
                                                  columns='channel',
                                                  values='value')

        buffer = io.BytesIO()
        if file_format == FileFormat.FEATHER:
            ts.to_feather(buffer)
            mime_type = 'application/feather'
        elif file_format == FileFormat.PARQUET:
            ts.to_parquet(buffer, index=True)
            mime_type = 'application/parquet'
        else:
            raise NotImplementedError(f"Unsupported file format: {file_format}")
        buffer.seek(0)
        data = buffer.read()

        first_timestamp = None if len(ts) == 0 else ts.index.get_level_values('timestamp').min()

        # Encode as a data-url
        b64_data = base64.b64encode(data).decode('utf-8')
        url = f"data:{mime_type};base64,{b64_data}"

        req = GetStatusRequestResponse(
            request_id=str(uuid.uuid4()),
            status_code=200,
            message="Success",
            requested_at=datetime.now(timezone.utc),
            completed_at=datetime.now(timezone.utc),
            result=GetStatusRequestResponse.GetDataRequestStatusCompletedResult(
                result_url=url,
                expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
                mime_type=mime_type,
                file_size=len(data),
                metadata=GetStatusRequestResponse.GetDataRequestStatusCompletedResult.ResultMetadata(
                    first_timestamp=first_timestamp
                )
            )
        )
        self._known_requests[req.request_id] = req
        return req

    def get_request_status(self, request_id: str) -> GetStatusRequestResponse:
        if request_id not in self._known_requests:
            raise MercutoHTTPException(status_code=404, message="Not Found")
        return self._known_requests[request_id]

    def insert_metric_samples(
        self,
        project: str,
        samples: Collection[MetricDataSample]
    ) -> None:
        if not samples:
            return

        # Ensure all channels are of type METRIC
        if not all(
            sample.channel in self._channels and (self._channels[sample.channel].classification == ChannelClassification.EVENT_METRIC or
                                                  self._channels[sample.channel].classification == ChannelClassification.PRIMARY_EVENT_AGGREGATE)
            and self._channels[sample.channel].project == project
            for sample in samples
        ):
            return

        df = pd.DataFrame([{
            'channel': s.channel,
            'timestamp': s.timestamp,
            'value': s.value,
            'event': s.event
        } for s in samples])
        df = df.set_index(['channel', 'timestamp'])

        to_concat: list[pd.DataFrame] = []
        if len(self._metric_buffer) > 0:
            to_concat.append(self._metric_buffer)
        if len(df) > 0:
            to_concat.append(df)

        self._metric_buffer = pd.concat(to_concat).sort_index()
        self._update_last_valid_samples()

    def insert_secondary_samples(
        self,
        project: str,
        samples: Collection[SecondaryDataSample]
    ) -> None:
        if not samples:
            return

        # Ensure all channels are of type SECONDARY
        if not all(
            sample.channel in self._channels and self._channels[sample.channel].classification == ChannelClassification.SECONDARY
            and self._channels[sample.channel].project == project
            for sample in samples
        ):
            return

        df = pd.DataFrame([{
            'channel': s.channel,
            'timestamp': s.timestamp,
            'value': s.value,
        } for s in samples])
        df = df.set_index(['channel', 'timestamp'])

        to_concat: list[pd.DataFrame] = []
        if len(self._secondary_and_primary_buffer) > 0:
            to_concat.append(self._secondary_and_primary_buffer)
        if len(df) > 0:
            to_concat.append(df)

        self._secondary_and_primary_buffer = pd.concat(to_concat).sort_index()
        self._update_last_valid_samples()

    def delete_metric_samples(self, project: str, event: str, channels: Optional[Collection[str]] = None) -> None:
        if channels is None:
            channels = [c.code for c in self._channels.values() if c.project == project]
        idx = self._metric_buffer.index

        mask = (
            idx.get_level_values('channel').isin(channels) &
            (self._metric_buffer['event'] == event)
        )
        self._metric_buffer = self._metric_buffer[~mask]
        self._update_last_valid_samples()

    def load_metric_samples(
        self,
        channels: Optional[Collection[str]] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        events: Optional[Collection[str]] = None,
        project: Optional[str] = None,
        limit: int = 100
    ) -> list[MetricDataSample]:
        if channels is None and project is None:
            raise ValueError("Must supply either channels or project.")

        if channels is None:
            channels = [c.code for c in self._channels.values() if c.project == project and c.classification in {
                ChannelClassification.EVENT_METRIC, ChannelClassification.PRIMARY_EVENT_AGGREGATE
            }]

        # Ensure all channels exist are of type METRIC or PRIMARY_EVENT_AGGREGATE
        if not all(
            ch in self._channels and self._channels[ch].classification in {
                ChannelClassification.EVENT_METRIC, ChannelClassification.PRIMARY_EVENT_AGGREGATE}
            for ch in channels
        ):
            return []

        idx = self._metric_buffer.index
        mask = (
            idx.get_level_values('channel').isin(channels) &
            (start_time is None or idx.get_level_values('timestamp') >= start_time) &
            (end_time is None or idx.get_level_values('timestamp') <= end_time) &
            (events is None or self._metric_buffer['event'].isin(events))
        )
        filtered = self._metric_buffer[mask]
        return [
            MetricDataSample(
                channel=channel,
                timestamp=timestamp,
                value=row['value'],
                event=row['event']
            )
            for (channel, timestamp), row in filtered.iterrows()
        ][:limit]

    def load_secondary_samples(
        self,
        channels: Collection[str],
        start_time: datetime,
        end_time: datetime,
        limit: int = 100
    ) -> list[SecondaryDataSample]:
        idx = self._secondary_and_primary_buffer.index
        mask = (
            idx.get_level_values('channel').isin(channels) &
            (idx.get_level_values('timestamp') >= start_time) &
            (idx.get_level_values('timestamp') <= end_time)
        )
        filtered = self._secondary_and_primary_buffer[mask]
        return [
            SecondaryDataSample(
                channel=channel,
                timestamp=timestamp,
                value=row['value']
            )
            for (channel, timestamp), row in filtered.iterrows()
        ][:limit]

    def list_datatables(self, project: str) -> list[Datatable]:
        return [dt for dt in self._datatables.values() if dt.project == project]

    def create_datatable(self, project: str, name: str, sampling_period: timedelta, column_labels: Collection[str]) -> Datatable:
        if sampling_period <= timedelta(seconds=1):
            classification = ChannelClassification.PRIMARY
        else:
            classification = ChannelClassification.SECONDARY
        channels = [self.create_channel(project=project, label=col, classification=classification,
                                        sampling_period=sampling_period) for col in column_labels]
        dt = Datatable(
            code=str(uuid.uuid4()),
            project=project,
            name=name,
            sampling_period=sampling_period,
            columns=[DatatableColumn(column_label=ch.label, channel=ch.code) for ch in channels],
            enabled=True
        )
        self._datatables[dt.code] = dt
        return dt

    def _dt_col_to_channel_code(self, dt: str, column_label: str) -> str:
        for col in self._datatables[dt].columns:
            if col.column_label == column_label:
                return col.channel
        raise ValueError(f"Column label '{column_label}' not found in datatable '{dt}'")

    def upload_file(self, project: str, datatable: str, file: str | bytes | TextIO | BinaryIO,
                    filename: Optional[str] = None,
                    timezone: Optional[str] = None) -> None:
        frame = pd.read_csv(file, header=1, skiprows=[2, 3],
                            usecols=None, sep=',', index_col=0, na_values=['NAN', '"NAN"'])
        frame.index = pd.to_datetime(frame.index, utc=False)
        assert isinstance(frame.index, pd.DatetimeIndex)
        if frame.index.tz is None:
            frame.index = frame.index.tz_convert(timezone)
        del frame['RECORD']

        # Drop unknown channels
        frame = frame[[col for col in frame.columns if col in {c.column_label for c in self._datatables[datatable].columns}]]

        # rename from label to code
        frame = frame.rename(columns=lambda x: self._dt_col_to_channel_code(datatable, x))

        frame = frame.melt([], tuple(frame.columns), ignore_index=False, var_name='channel',
                           value_name='value').sort_index()
        frame.index.name = 'timestamp'
        frame = frame.reset_index().set_index(['channel', 'timestamp'])
        self._secondary_and_primary_buffer = pd.concat([self._secondary_and_primary_buffer, frame]).sort_index()
        self._update_last_valid_samples()

    def get_latest_samples(self, project: str, include_primary: bool = True) -> list[LatestDataSample]:
        if include_primary:
            channels = [c.code for c in self._channels.values() if c.project == project]
        else:
            channels = [ch for ch in channels if self._channels[ch].classification !=
                        ChannelClassification.PRIMARY and self._channels[ch].project == project]

        out: list[LatestDataSample] = []

        # Get the last timestamp and value for each channel in the secondary_and_primary_buffer
        latest = self._secondary_and_primary_buffer.groupby(level='channel').last()
        for (channel, timestamp), row in latest.iterrows():
            if channel not in channels:
                continue
            out.append(LatestDataSample(channel=channel,
                                        timestamp=timestamp,
                                        value=row['value']))

        latest = self._metric_buffer.groupby(level='channel').last()
        for (channel, timestamp), row in latest.iterrows():
            if channel not in channels:
                continue
            out.append(LatestDataSample(channel=channel,
                                        timestamp=timestamp,
                                        value=row['value']))
        return out

    def get_unit(self, code: str) -> Optional[Units]:
        return self._units.get(code)

    def list_units(self) -> list[Units]:
        return list(self._units.values())

    def create_unit(self, name: str, unit: str) -> Units:
        code = str(uuid.uuid4())
        unit_obj = Units(code=code, name=name, unit=unit)
        self._units[code] = unit_obj
        return unit_obj
